#!/usr/bin/python
# groups.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (groups.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: groups


BitDust Groups are intended to organize isolated private data flows between multiple users.

Owner of the group (customer) first creates a new "group key" that will be used to
protect data flows between participants of the group.

To add a new member to the group he just need to share the private part of that key
with the trusted person - that is implemented in `group_access_donor()` state machine.
Remote user once received the key will recognize that the key is actually a "group key"
(alias of the key must begins with `group_`) and will start a new instance
of `group_access_coordinator()` state machine which suppose to connect him to
other participants of the group.

In order to run data flows between multiple users a new role was introduced which is
called `postman`. All group communications are done via postmans - those are holding
the queues and group participants are connected to the queues as "consumers" and "producers".
Postmans are running on same devices where suppliers of the group creator are hoster.
That way all group participants always know where to connect to because list of suppliers
is already available in the DHT.

The postman is not able to read message content, but only validate the signature
and verify that participants are authorized.

Each group participant selects only one postman to publish messages to the group, but it always
subscribes to all of the known postmans to listen for incoming messages. In case selected postman goes offline
the group participant will switch to the next known postman in the list.

It is possible that two group participants simultaneously sent two different messages to the group with same
sequence ID. That creates a conflict in the sequence of messages delivered to the group memebers.
It is solved directly on each group participant node.

In short group is:
    1. group key created by "group owner" and shared to participants
    2. messages queues are served by postmans
    3. consumers and producers subscribed to the queues and able to send/receive messages

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from twisted.internet.defer import DeferredList

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import utime
from bitdust.lib import jsn

from bitdust.system import local_fs
from bitdust.system import bpio

from bitdust.contacts import contactsdb

from bitdust.main import settings

from bitdust.crypt import key
from bitdust.crypt import my_keys

from bitdust.access import key_ring

from bitdust.userid import id_url
from bitdust.userid import global_id

#------------------------------------------------------------------------------

REQUIRED_BROKERS_COUNT = 3

_ActiveGroups = {}

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'groups.init')
    load_groups()


def shutdown():
    unload_groups()
    if _Debug:
        lg.out(_DebugLevel, 'groups.shutdown')


#------------------------------------------------------------------------------


def active_groups():
    global _ActiveGroups
    return _ActiveGroups


#------------------------------------------------------------------------------


def is_group_exist(group_key_id):
    group_key_id = my_keys.latest_key_id(group_key_id)
    return group_key_id in active_groups()


def is_group_stored(group_key_id):
    service_dir = settings.ServiceDir('service_private_groups')
    groups_dir = os.path.join(service_dir, 'groups')
    group_info_path = os.path.join(groups_dir, group_key_id)
    return os.path.isfile(group_info_path)


def generate_group_key(creator_id=None, label=None, key_size=4096, group_alias=None):
    group_key_id = None
    if group_alias:
        group_key_id = my_keys.make_key_id(alias=group_alias, creator_glob_id=creator_id)
        if my_keys.is_key_registered(group_key_id):
            return my_keys.latest_key_id(group_key_id)
    else:
        while True:
            random_sample = os.urandom(24)
            group_alias = 'group_%s' % strng.to_text(key.HashMD5(random_sample, hexdigest=True))
            group_key_id = my_keys.make_key_id(alias=group_alias, creator_glob_id=creator_id)
            if my_keys.is_key_registered(group_key_id):
                continue
            break
    if not label:
        label = 'group%s' % utime.make_timestamp()
    my_keys.generate_key(key_id=group_key_id, label=label, key_size=key_size)
    my_keys.sign_key(key_id=group_key_id, save=True)
    if _Debug:
        lg.args(_DebugLevel, group_key_id=group_key_id, group_alias=group_alias, creator_id=creator_id, label=label)
    return group_key_id


def set_group_info(group_key_id, group_info=None):
    group_key_id = my_keys.latest_key_id(group_key_id)
    if not group_info:
        group_info = {
            'last_sequence_id': -1,
            'active': False,
        }
    active_groups()[group_key_id] = group_info
    return True


def create_new_group(label, creator_id=None, key_size=2048, group_alias=None, with_group_info=True):
    if _Debug:
        lg.args(_DebugLevel, label=label, creator_id=creator_id, key_size=key_size, group_alias=group_alias)
    group_key_id = generate_group_key(creator_id=creator_id, label=label, key_size=key_size, group_alias=group_alias)
    if with_group_info:
        set_group_info(group_key_id, group_info={'last_sequence_id': -1, 'active': False})
        save_group_info(group_key_id)
    return group_key_id


#------------------------------------------------------------------------------


def send_group_pub_key_to_suppliers(group_key_id):
    l = []
    for supplier_idurl in contactsdb.suppliers():
        if supplier_idurl:
            d = key_ring.transfer_key(group_key_id, supplier_idurl, include_private=False, include_signature=False)
            d.addCallback(lg.cb, debug=_Debug, debug_level=_DebugLevel, method='groups.write_group_key_to_suppliers')
            d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='groups.write_group_key_to_suppliers')
            # TODO: build some kind of retry mechanism - in case of a particular supplier did not receive the key
            # it must be some process with each supplier that first verifies a list of my public keys supplier currently possess
            # and then transfer the missing keys or send a note to erase "unused" keys to be able to cleanup old keys
            l.append(d)
    return DeferredList(l, consumeErrors=True)


#------------------------------------------------------------------------------


def load_groups():
    loaded_groups = 0
    service_dir = settings.ServiceDir('service_private_groups')
    groups_dir = os.path.join(service_dir, 'groups')
    if not os.path.isdir(groups_dir):
        bpio._dirs_make(groups_dir)
    for group_key_id in os.listdir(groups_dir):
        latest_group_key_id = my_keys.latest_key_id(group_key_id)
        latest_group_path = os.path.join(groups_dir, latest_group_key_id)
        if latest_group_key_id != group_key_id:
            lg.info('going to rename rotated group key: %r -> %r' % (group_key_id, latest_group_key_id))
            old_group_path = os.path.join(groups_dir, group_key_id)
            try:
                os.rename(old_group_path, latest_group_path)
            except:
                lg.exc()
                continue
        latest_group_info = jsn.loads_text(local_fs.ReadTextFile(latest_group_path))
        if not latest_group_info:
            lg.err('was not able to load group info from %r' % latest_group_path)
            continue
        active_groups()[latest_group_key_id] = latest_group_info
        loaded_groups += 1
    if _Debug:
        lg.args(_DebugLevel, loaded_groups=loaded_groups)


def unload_groups():
    active_groups().clear()
    if _Debug:
        lg.dbg(_DebugLevel, 'known groups and brokers are erased')


#------------------------------------------------------------------------------


def save_group_info(group_key_id):
    group_key_id = my_keys.latest_key_id(group_key_id)
    if not is_group_exist(group_key_id):
        lg.warn('group %r is not known' % group_key_id)
        return False
    group_info = active_groups()[group_key_id]
    service_dir = settings.ServiceDir('service_private_groups')
    groups_dir = os.path.join(service_dir, 'groups')
    group_info_path = os.path.join(groups_dir, group_key_id)
    if not os.path.isdir(groups_dir):
        bpio._dirs_make(groups_dir)
    ret = local_fs.WriteTextFile(group_info_path, jsn.dumps(group_info))
    if _Debug:
        lg.args(_DebugLevel, group_key_id=group_key_id, group_info_path=group_info_path, ret=ret)
    return ret


def erase_group_info(group_key_id):
    if not is_group_exist(group_key_id):
        lg.warn('group %r is not known' % group_key_id)
        return False
    service_dir = settings.ServiceDir('service_private_groups')
    groups_dir = os.path.join(service_dir, 'groups')
    group_info_path = os.path.join(groups_dir, group_key_id)
    if not os.path.isfile(group_info_path):
        return False
    os.remove(group_info_path)
    if _Debug:
        lg.args(_DebugLevel, group_key_id=group_key_id, group_info_path=group_info_path)
    return True


def read_group_info(group_key_id):
    service_dir = settings.ServiceDir('service_private_groups')
    groups_dir = os.path.join(service_dir, 'groups')
    group_info_path = os.path.join(groups_dir, group_key_id)
    if not os.path.isfile(group_info_path):
        return None
    group_info = jsn.loads_text(local_fs.ReadTextFile(group_info_path))
    return group_info


#------------------------------------------------------------------------------


def get_last_sequence_id(group_key_id):
    group_key_id = my_keys.latest_key_id(group_key_id)
    if not is_group_exist(group_key_id):
        return -1
    return active_groups()[group_key_id]['last_sequence_id']


def set_last_sequence_id(group_key_id, last_sequence_id):
    group_key_id = my_keys.latest_key_id(group_key_id)
    if not is_group_exist(group_key_id):
        lg.warn('group %r is not known' % group_key_id)
        return False
    active_groups()[group_key_id]['last_sequence_id'] = last_sequence_id
    return True


#------------------------------------------------------------------------------


def is_group_active(group_key_id):
    group_key_id = my_keys.latest_key_id(group_key_id)
    if not is_group_exist(group_key_id):
        return False
    return active_groups()[group_key_id]['active']


def set_group_active(group_key_id, value):
    group_key_id = my_keys.latest_key_id(group_key_id)
    if not is_group_exist(group_key_id):
        lg.warn('group %r is not known' % group_key_id)
        return False
    old_value = active_groups()[group_key_id]['active']
    active_groups()[group_key_id]['active'] = value
    if old_value != value:
        lg.info('group %r "active" status changed: %r -> %r' % (group_key_id, old_value, value))
    return True


#------------------------------------------------------------------------------


def on_identity_url_changed(evt):
    from bitdust.access import group_participant
    service_dir = settings.ServiceDir('service_private_groups')
    groups_dir = os.path.join(service_dir, 'groups')
    old_idurl = id_url.field(evt.data['old_idurl'])
    active_group_keys = list(active_groups())
    to_be_reconnected = []
    for group_key_id in active_group_keys:
        if not group_key_id:
            continue
        group_creator_idurl = global_id.glob2idurl(group_key_id)
        if id_url.is_the_same(group_creator_idurl, old_idurl):
            old_group_path = os.path.join(groups_dir, group_key_id)
            latest_group_key_id = my_keys.latest_key_id(group_key_id)
            latest_group_path = os.path.join(groups_dir, latest_group_key_id)
            lg.info('going to rename rotated group file: %r -> %r' % (old_group_path, latest_group_path))
            if os.path.isfile(old_group_path):
                try:
                    os.rename(old_group_path, latest_group_path)
                except:
                    lg.exc()
                    continue
            else:
                lg.warn('key file %r was not found, key was not renamed' % old_group_path)
            active_groups()[latest_group_key_id] = active_groups().pop(group_key_id)
            group_participant.rotate_active_group_participant(group_key_id, latest_group_key_id)
        gp = group_participant.get_active_group_participant(group_key_id)
        if gp and gp.active_supplier_idurl and id_url.is_the_same(old_idurl, gp.active_supplier_idurl):
            lg.info('active supplier %r IDURL is rotated, going to reconnect %r' % (old_idurl, gp))
            if group_key_id not in to_be_reconnected:
                to_be_reconnected.append(group_key_id)
    if _Debug:
        lg.args(_DebugLevel, to_be_reconnected=to_be_reconnected)
    for group_key_id in to_be_reconnected:
        gp = group_participant.get_active_group_participant(group_key_id)
        if gp:
            gp.automat('reconnect')



#!/usr/bin/python
# index_synchronizer.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (index_synchronizer.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: index_synchronizer.

.. role:: red

.. raw:: html

    <a href="https://bitdust.io/automats/index_synchronizer/index_synchronizer.png" target="_blank">
    <img src="https://bitdust.io/automats/index_synchronizer/index_synchronizer.png" style="max-width:100%;">
    </a>


Here is a state machine ``index_synchronizer()``, it is aimed to synchronize
local backup index database with remote suppliers.

This allows to restore the index file (with all your backup IDs and files and folders names)
from your suppliers in case of data lost.

The purpose of index_synchronizer() automat is to sync users's backup database to remote computers.
In case of loss of all local data - backup database will be restored from the suppliers.

The database includes the list of folders to be backed up, optional schedule for the backups,
and most importantly, a list of already created backups by its ID.

Thus if the user has made recovery of his account and restore the backup database -
he can recover his data from remote machines by recognized backup ID.

Every time any local change was made to the database it must synchronized with remote copies.

At first step, index_synchronizer() requests all remote copies of the database from suppliers.
When new file arrives from supplier, "backup_control" starts a validation against
current local index file and update local copy if required.
On next step index_synchronizer() sends a latest version of index file to all suppliers to hold.

The backup_monitor() machine should be restarted every one hour
or every time when your files were changed.
It sends "restart" event to index_synchronizer() to synchronize index file.


BitDust index_synchronizer() Automat

EVENTS:
    * :red:`all-acked`
    * :red:`all-responded`
    * :red:`index-file-received`
    * :red:`init`
    * :red:`instant`
    * :red:`pull`
    * :red:`push`
    * :red:`shutdown`
    * :red:`timer-10sec`
    * :red:`timer-15sec`
    * :red:`timer-1min`
    * :red:`timer-5min`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import time

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import nameurl

from bitdust.p2p import commands
from bitdust.p2p import online_status
from bitdust.p2p import p2p_service
from bitdust.p2p import propagate

from bitdust.system import bpio

from bitdust.userid import my_id
from bitdust.userid import global_id

from bitdust.contacts import contactsdb

from bitdust.main import settings
from bitdust.main import events

from bitdust.crypt import encrypted
from bitdust.crypt import signed
from bitdust.crypt import key

from bitdust.transport import packet_out

from bitdust.services import driver

from bitdust.customer import supplier_connector

#------------------------------------------------------------------------------

_IndexSynchronizer = None

#------------------------------------------------------------------------------


def is_synchronized():
    if not A():
        return False
    if A().state == 'IN_SYNC!':
        return True
    if A().state in ['REQUEST?', 'SENDING']:
        if A().last_time_in_sync > 0 and time.time() - A().last_time_in_sync < 30:
            return True
    return False


def is_synchronizing():
    if not A():
        return False
    return A().state in ['REQUEST?', 'SENDING']


#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _IndexSynchronizer
    if event is None and not args:
        return _IndexSynchronizer
    if _IndexSynchronizer is None:
        _IndexSynchronizer = IndexSynchronizer(
            name='index_synchronizer',
            state='AT_STARTUP',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _IndexSynchronizer.automat(event, *args, **kwargs)
    return _IndexSynchronizer


#------------------------------------------------------------------------------


class IndexSynchronizer(automat.Automat):

    """
    This class implements all the functionality of the ``index_synchronizer()``
    state machine.
    """

    fast = False

    timers = {
        'timer-1min': (60, ['NO_INFO']),
        'timer-10sec': (10.0, ['REQUEST?']),
        'timer-15sec': (15.0, ['REQUEST?', 'SENDING']),
        'timer-5min': (300, ['IN_SYNC!']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of index_synchronizer() machine.
        """
        self.latest_supplier_revision = -1
        self.current_local_revision = -1
        self.requesting_suppliers = set()
        self.requests_packets_sent = []
        self.requested_suppliers_number = 0
        self.sending_suppliers = set()
        self.sent_suppliers_number = 0
        self.outgoing_packets_ids = []
        self.last_time_in_sync = -1
        self.PushAgain = False

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Method to catch the moment when index_synchronizer() state were
        changed.
        """
        if newstate == 'IN_SYNC!':
            if A().last_time_in_sync > 0 and time.time() - A().last_time_in_sync < 30:
                if _Debug:
                    lg.dbg(_DebugLevel, 'backup index already synchronized %r seconds ago' % (time.time() - A().last_time_in_sync))
            else:
                if _Debug:
                    lg.dbg(_DebugLevel, 'backup index just synchronized, sending "my-backup-index-synchronized" event')
                events.send('my-backup-index-synchronized', data={})
            self.last_time_in_sync = time.time()
            if self.PushAgain:
                reactor.callLater(0, self.automat, 'instant')  # @UndefinedVariable
        if newstate == 'NO_INFO' and oldstate in [
            'REQUEST?',
            'SENDING',
        ]:
            events.send('my-backup-index-out-of-sync', data={})
        if newstate == 'NO_INFO':
            self.last_time_in_sync = -1

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'NO_INFO'
                self.doInit(*args, **kwargs)
        #---IN_SYNC!---
        elif self.state == 'IN_SYNC!':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'pull' or event == 'timer-5min':
                self.state = 'REQUEST?'
                self.doSuppliersRequestIndexFile(*args, **kwargs)
            elif event == 'push' or (event == 'instant' and self.PushAgain):
                self.state = 'SENDING'
                self.doSuppliersSendIndexFile(*args, **kwargs)
                self.PushAgain = False
                self.PullAgain = False
        #---REQUEST?---
        elif self.state == 'REQUEST?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doCancelRequests(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif (event == 'all-responded' or (event == 'timer-15sec' and self.isSomeResponded(*args, **kwargs))) and self.isVersionChanged(*args, **kwargs):
                self.state = 'SENDING'
                self.doCancelRequests(*args, **kwargs)
                self.doSuppliersSendIndexFile(*args, **kwargs)
                self.PushAgain = False
                self.PullAgain = False
            elif event == 'index-file-received':
                self.doCheckVersion(*args, **kwargs)
            elif (event == 'all-responded' or (event == 'timer-15sec' and self.isSomeResponded(*args, **kwargs))) and not self.isVersionChanged(*args, **kwargs):
                self.state = 'IN_SYNC!'
                self.doCancelRequests(*args, **kwargs)
            elif event == 'timer-10sec' and not self.isSomeResponded(*args, **kwargs) and self.isAllTimedOut(*args, **kwargs):
                self.state = 'NO_INFO'
                self.doCancelRequests(*args, **kwargs)
        #---SENDING---
        elif self.state == 'SENDING':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doCancelSendings(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'push':
                self.PushAgain = True
            elif event == 'timer-15sec' and not self.isSomeAcked(*args, **kwargs) and not self.PullAgain:
                self.state = 'NO_INFO'
                self.doCancelSendings(*args, **kwargs)
            elif (event == 'all-acked' or (event == 'timer-15sec' and self.isSomeAcked(*args, **kwargs))) and not self.PullAgain:
                self.state = 'IN_SYNC!'
                self.doCancelSendings(*args, **kwargs)
            elif event == 'pull':
                self.PullAgain = True
            elif (event == 'all-acked' or event == 'timer-15sec') and self.PullAgain:
                self.state = 'REQUEST?'
                self.doCancelSendings(*args, **kwargs)
                self.doSuppliersRequestIndexFile(*args, **kwargs)
                self.PullAgain = False
        #---NO_INFO---
        elif self.state == 'NO_INFO':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'push' or event == 'pull' or event == 'timer-1min':
                self.state = 'REQUEST?'
                self.doSuppliersRequestIndexFile(*args, **kwargs)
                self.PushAgain = False
                self.PullAgain = False
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isSomeAcked(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.sending_suppliers) < self.sent_suppliers_number

    def isSomeResponded(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.requesting_suppliers) < self.requested_suppliers_number

    def isVersionChanged(self, *args, **kwargs):
        """
        Condition method.
        """
        if self.current_local_revision < 0:
            # no info about current local version - assume version was changed
            return True
        return self.current_local_revision != self.latest_supplier_revision

    def isAllTimedOut(self, *args, **kwargs):
        """
        Condition method.
        """
        for packetID, supplierIDURL in self.requests_packets_sent:
            pkts_out = packet_out.search_many(
                command=commands.Retrieve(),
                remote_idurl=supplierIDURL,
                packet_id=packetID,
            )
            if pkts_out:
                return False
        return True

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        self.ping_required = False

    def doSuppliersRequestIndexFile(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'index_synchronizer.doSuppliersRequestIndexFile')
        if driver.is_on('service_backups'):
            from bitdust.storage import backup_fs
            self.current_local_revision = backup_fs.revision()
        else:
            self.current_local_revision = -1
        self.latest_supplier_revision = -1
        self.requesting_suppliers.clear()
        self.requested_suppliers_number = 0
        self.requests_packets_sent = []
        if self.ping_required:
            propagate.ping_suppliers().addBoth(self._do_retrieve)
            self.ping_required = False
        else:
            self._do_retrieve()

    def doSuppliersSendIndexFile(self, *args, **kwargs):
        """
        Action method.
        """
        packetID = global_id.MakeGlobalID(
            customer=my_id.getGlobalID(key_alias='master'),
            path=settings.BackupIndexFileName(),
        )
        self.sending_suppliers.clear()
        self.outgoing_packets_ids = []
        self.sent_suppliers_number = 0
        localID = my_id.getIDURL()
        data = bpio.ReadBinaryFile(settings.BackupIndexFilePath())
        b = encrypted.Block(
            CreatorID=localID,
            BackupID=packetID,
            BlockNumber=0,
            SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
            SessionKeyType=key.SessionKeyType(),
            LastBlock=True,
            Data=data,
        )
        Payload = b.Serialize()
        if _Debug:
            lg.args(_DebugLevel, pid=packetID, sz=len(data), payload=len(Payload), length=b.Length)
        for supplier_idurl in contactsdb.suppliers():
            if not supplier_idurl:
                continue
            sc = supplier_connector.by_idurl(supplier_idurl)
            if sc is None or sc.state != 'CONNECTED':
                continue
            if online_status.isOffline(supplier_idurl):
                continue
            newpacket, pkt_out = p2p_service.SendData(
                raw_data=Payload,
                ownerID=localID,
                creatorID=localID,
                remoteID=supplier_idurl,
                packetID=packetID,
                callbacks={
                    commands.Ack(): self._on_supplier_acked,
                    commands.Fail(): self._on_supplier_acked,
                },
            )
            if pkt_out:
                self.sending_suppliers.add(supplier_idurl)
                self.sent_suppliers_number += 1
                if newpacket.PacketID not in self.outgoing_packets_ids:
                    self.outgoing_packets_ids.append(newpacket.PacketID)
            if _Debug:
                lg.out(_DebugLevel, '    %s sending to %s' % (newpacket, nameurl.GetName(supplier_idurl)))

    def doCancelSendings(self, *args, **kwargs):
        """
        Action method.
        """
        for packet_id in self.outgoing_packets_ids:
            packetsToCancel = packet_out.search_by_packet_id(packet_id)
            for pkt_out in packetsToCancel:
                if pkt_out.outpacket.Command == commands.Data():
                    pkt_out.automat('cancel')

    def doCancelRequests(self, *args, **kwargs):
        """
        Action method.
        """


#         packetID = global_id.MakeGlobalID(
#             customer=my_id.getGlobalID(key_alias='master'),
#             path=settings.BackupIndexFileName(),
#         )
#         from bitdust.transport import packet_out
#         packetsToCancel = packet_out.search_by_packet_id(packetID)
#         for pkt_out in packetsToCancel:
#             if pkt_out.outpacket.Command == commands.Retrieve():
#                 lg.warn('sending "cancel" to %s addressed to %s from index_synchronizer' % (
#                     pkt_out, pkt_out.remote_idurl, ))
#                 pkt_out.automat('cancel')

    def doCheckVersion(self, *args, **kwargs):
        """
        Action method.
        """
        _, supplier_revision = args[0]
        if supplier_revision > self.latest_supplier_revision:
            self.latest_supplier_revision = supplier_revision

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()
        global _IndexSynchronizer
        del _IndexSynchronizer
        _IndexSynchronizer = None

    def _on_supplier_response(self, newpacket, info):
        wrapped_packet = signed.Unserialize(newpacket.Payload)
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket, wrapped_packet=wrapped_packet)
        if not wrapped_packet or not wrapped_packet.Valid():
            lg.err('incoming Data() is not valid')
            return
        supplier_idurl = wrapped_packet.RemoteID
        from bitdust.storage import backup_control
        supplier_revision = backup_control.IncomingSupplierBackupIndex(wrapped_packet)
        self.requesting_suppliers.discard(supplier_idurl)
        if supplier_revision is not None:
            reactor.callLater(0, self.automat, 'index-file-received', (newpacket, supplier_revision))  # @UndefinedVariable
        if _Debug:
            lg.out(_DebugLevel, 'index_synchronizer._on_supplier_response %s from %r, rev:%s, pending: %d, total: %d' % (newpacket, supplier_idurl, supplier_revision, len(self.requesting_suppliers), self.requested_suppliers_number))
        if len(self.requesting_suppliers) == 0:
            reactor.callLater(0, self.automat, 'all-responded')  # @UndefinedVariable

    def _on_supplier_fail(self, newpacket, info):
        if _Debug:
            lg.args(_DebugLevel, newpacket=newpacket)
        supplier_idurl = newpacket.CreatorID
        self.requesting_suppliers.discard(supplier_idurl)
        if _Debug:
            lg.out(_DebugLevel, 'index_synchronizer._on_supplier_fail %s from %r, pending: %d, total: %d' % (newpacket, supplier_idurl, len(self.requesting_suppliers), self.requested_suppliers_number))
        if len(self.requesting_suppliers) == 0:
            reactor.callLater(0, self.automat, 'all-responded')  # @UndefinedVariable

    def _on_supplier_acked(self, newpacket, info):
        self.sending_suppliers.discard(newpacket.OwnerID)
        # if newpacket.PacketID in self.outgoing_packets_ids:
        #     self.outgoing_packets_ids.remove(newpacket.PacketID)
        sc = supplier_connector.by_idurl(newpacket.OwnerID)
        if sc:
            sc.automat(newpacket.Command.lower(), newpacket)
        else:
            lg.warn('did not found supplier connector for %r' % newpacket.OwnerID)
        if _Debug:
            lg.out(_DebugLevel, 'index_synchronizer._on_supplier_acked %s, pending: %d, total: %d' % (newpacket, len(self.sending_suppliers), self.sent_suppliers_number))
        if len(self.sending_suppliers) == 0:
            reactor.callLater(0, self.automat, 'all-acked')  # @UndefinedVariable

    def _do_retrieve(self, x=None):
        packetID = global_id.MakeGlobalID(
            customer=my_id.getGlobalID(key_alias='master'),
            path=settings.BackupIndexFileName(),
        )
        localID = my_id.getIDURL()
        for supplier_idurl in contactsdb.suppliers():
            if not supplier_idurl:
                continue
            sc = supplier_connector.by_idurl(supplier_idurl)
            if sc is None or sc.state != 'CONNECTED':
                continue
            if online_status.isOffline(supplier_idurl):
                continue
            pkt_out = p2p_service.SendRetreive(
                ownerID=localID,
                creatorID=localID,
                packetID=packetID,
                remoteID=supplier_idurl,
                response_timeout=settings.P2PTimeOut(),
                callbacks={
                    commands.Data(): self._on_supplier_response,
                    commands.Fail(): self._on_supplier_fail,
                },
            )
            if pkt_out:
                self.requesting_suppliers.add(supplier_idurl)
                self.requested_suppliers_number += 1
                self.requests_packets_sent.append((packetID, supplier_idurl))
            if _Debug:
                lg.out(_DebugLevel, '    %s sending to %s' % (pkt_out, nameurl.GetName(supplier_idurl)))



"""
A Tornado based wallet server for Bismuth, with websocket server.

EggdraSyl -October 2018

pip3 install -r requirements.txt
"""

# TODO: background status tasks

# TODO: limit clients #

# TODO: send copies of txs directly to pool nodes? Ok if wallet servers are whitelisted.

# TODO: route with a key to show last logs

# TODO: wstatusget , total slots and connected clients.

import logging
import json
import os
import sys
import time
from logging.handlers import RotatingFileHandler

import aioprocessing
import psutil
import tornado.gen
import tornado.log
import tornado.util
from tornado.options import define, options
import tornado.web
import tornado.httpserver
import tornado.websocket

# Bismuth specific modules
import modules.config as config
# from modules.helpers import *
from modules.sqlitebase import SqliteBase
from modules.ledgerbase import LedgerBase
from modules.node_interface import NodeInterface

__version__ = '0.0.20'

NODE_INTERFACE: NodeInterface = None


class ChannelHandler(tornado.websocket.WebSocketHandler):
    """
    Handler for a websocket channel
    """

    client_id_counter = 0

    @classmethod
    def urls(cls):
        return [
            (r'/web-socket/', cls, {}),  # Route/Handler/kwargs
        ]

    def initialize(self):
        pass

    def open(self):
        """
        Client opens a websocket
        """
        global access_log
        global app_log
        #client id could be remote_ip/port instead, but need to handle proxies/ports
        ChannelHandler.client_id_counter += 1
        self.client_id = f'{ChannelHandler.client_id_counter}({self.request.remote_ip})'
        self.message_id = 0
        access_log.info(f'open {self.client_id}')

    async def send_ko(self, reason):
        await self.write_message('["Ko", "{}"]'.format(reason))

    async def on_message(self, message):
        """
        Message received on channel
        """
        self.message_id += 1
        app_log.info(f'Message[{self.message_id}] from {self.client_id} {message}')
        message = json.loads(message)
        # TODO: check with message[0] that we have the right number of params
        params_count = NODE_INTERFACE.param_count_of(message[0], ['none'])
        if params_count < 0:
            await self.send_ko('Unknown command')
            return
        start_time = time.time()
        # string, or dict that will be json encoded
        res = await NODE_INTERFACE.call_user(message)
        time_taken = time.time() - start_time
        app_log.info(f'Response[{self.message_id}] for {self.client_id} took {time_taken:.3f}s')

        await self.write_message(json.dumps(res))

    def on_close(self):
        """
        Channel is closed
        """
        if self.close_code:
            access_log.info('close, code {} reason {}'.format(self.close_code, self.close_reason))

    def check_origin(self, origin):
        """
        Override the origin check if needed
        """
        return True


async def getrights(ip):
    try:
        result = ['none']
        if ip in CONFIG.allowed:
            result.append('admin')
        return result
    except Exception as e:
        app_log.warning('Error getrights {}'.format(str(e)))


def start_server(port):
    global NODE_INTERFACE

    io_loop = tornado.ioloop.IOLoop.current()
    mempool = SqliteBase(options.verbose, db_path=CONFIG.mempool_path.replace('mempool.db', ''), db_name='mempool.db', app_log=app_log)
    db_name = 'ledger.db'
    if CONFIG.testnet:
        db_name = 'test.db'
    ledger = LedgerBase(options.verbose, db_path=CONFIG.db_path + '/', db_name=db_name, app_log=app_log)
    try:
        # Force a db connection attempt and updates db version of ledger
        _ = io_loop.run_sync(ledger.check_db_version, 30)
    except Exception as e:
        app_log.error("Can't connect to ledger: {}".format(e))
        return

    NODE_INTERFACE = NodeInterface(mempool, ledger, CONFIG, app_log=app_log)

    app = tornado.web.Application(ChannelHandler.urls())

    # Setup HTTP Server
    http_server = tornado.httpserver.HTTPServer(app)
    http_server.listen(port)
    # http_server.listen(port, LISTEN_ADDRESS)

    # Start IO/Event loop
    try:
        io_loop.start()
    except KeyboardInterrupt:
        stop_event.set()
        io_loop.stop()
        app_log.info('exited cleanly')


if __name__ == '__main__':
    global app_log
    global access_log
    global stop_event
    global start_time
    global process

    CONFIG = config.Get()
    CONFIG.read()
    #
    is_testnet = CONFIG.testnet
    PORT = CONFIG.websocket_port
    MAX_CLIENTS = CONFIG.max_clients

    define('port', default=PORT, help='port to listen on')
    define('verbose', default=False, help='verbose')
    options.parse_command_line()

    # TODO
    """
    if CONFIG.mempool_ram_conf:
        print("Incompatible setting detected.")
        print("Please edit config.txt, set mempool_ram_conf=False and restart node")
        sys.exit()
    """

    # TODO: print settings

    if not os.path.isfile(CONFIG.mempool_path):
        print('mempool.db not found at {}'.format(CONFIG.mempool_path))
        print('Please edit config.txt, check mempool_ram_conf=False and restart node.')
        sys.exit()

    start_time = time.time()

    stop_event = aioprocessing.AioEvent()  # Event()

    lock = aioprocessing.AioLock()

    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    # formatter = logging.Formatter('%(asctime)s %(funcName)s(%(lineno)d) %(message)s')
    # ch.setFormatter(formatter)
    app_log = logging.getLogger('tornado.application')
    tornado.log.enable_pretty_logging()
    # app_log.addHandler(ch)
    logfile = os.path.abspath('websocket_app.log')
    # Rotate log after reaching 512K, keep 5 old copies.
    rotateHandler = RotatingFileHandler(logfile, 'a', 512*1024, 10)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    rotateHandler.setFormatter(formatter)
    app_log.addHandler(rotateHandler)

    access_log = logging.getLogger('tornado.access')
    tornado.log.enable_pretty_logging()
    logfile2 = os.path.abspath('websocket_access.log')
    rotateHandler2 = RotatingFileHandler(logfile2, 'a', 512*1024, 10)
    formatter2 = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    rotateHandler2.setFormatter(formatter2)
    access_log.addHandler(rotateHandler2)

    app_log.warning('Testnet: {}'.format(is_testnet))

    if os.name == 'posix':
        process = psutil.Process()
        try:
            limit = process.rlimit(psutil.RLIMIT_NOFILE)
        except:
            limit = (1024, -1)
        app_log.info('OS File limits {}, {}'.format(limit[0], limit[1]))
        if limit[0] < 1024:
            app_log.error('Too small ulimit, please tune your system.')
            sys.exit()
    else:
        process = None

    app_log.info('Websocket Server {} Starting on port {}.'.format(__version__, options.port))

    start_server(options.port)



#!/usr/bin/python
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (commands.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: commands.

This module describes all commands in the BitDust p2p communication protocol.
The command is stored as a string in the packet.Command field.
If all commands are repeatable, then sequence numbers are not so critical,
though we would want date so time for replay trouble was limited.
If backups are write, read, delete (and never write again), then replay
is not an issue here and we use PacketID that identifies data.
So we want things like "replace supplier Vincecate"  not "replace supplier 5"
where seeing command an extra time would not hurt.

These are the valid values for the command field of a packet:
    - Data/Ack/                 (if customer sends you data you store it) (could be parity packet too)
    - Retrieve/Data|Fail        (data packets returned exactly as is with our signature)
    - ListFiles/Files           (ask supplier to list backup IDs he knows about for us)
    - Identity/Ack              (ping/pong packets)
    - Message/Ack               (chat)
    - Coin/Ack                  (for contracts publishing/management)
"""

#------------------------------------------------------------------------------

P2PCommandAcks = None
RelayCommands = None

#------------------------------------------------------------------------------


def init():
    """
    Initialize a list of valid p2p commands.
    """
    global P2PCommandAcks
    global RelayCommands
    P2PCommandAcks = {}
    RelayCommands = set()
    # No Ack for Ack
    P2PCommandAcks[Ack()] = []
    # No Ack for Fail
    P2PCommandAcks[Fail()] = []
    # Ack Data with Ack or Fail unless it is our data coming back (would be only after Retrieve)
    P2PCommandAcks[Data()] = [
        Ack(),
        Fail(),
    ]
    # Ack Retrieve with Data or Fail
    P2PCommandAcks[Retrieve()] = [
        Data(),
        Fail(),
    ]
    # Ack ListFiles with Files
    P2PCommandAcks[ListFiles()] = [
        Files(),
        Fail(),
    ]
    # Ack Files with Ack or Fail
    P2PCommandAcks[Files()] = [
        Ack(),
        Fail(),
    ]
    # If identity comes in and no interested party then transport sends an Ack
    P2PCommandAcks[Identity()] = [
        Ack(),
        Fail(),
    ]
    # Ack with Ack (maybe should be Files)
    P2PCommandAcks[DeleteFile()] = [
        Ack(),
        Fail(),
    ]
    # Ack with Ack (maybe should be Files)
    P2PCommandAcks[DeleteBackup()] = [
        Ack(),
        Fail(),
    ]
    # Ack with Ack or Fail, but also Message() packet may have no ack when sending back archived messages
    P2PCommandAcks[Message()] = [
        Ack(),
        Fail(),
    ]
    # Ack with Ack or Fail
    P2PCommandAcks[Receipt()] = [
        Ack(),
        Fail(),
    ]
    P2PCommandAcks[Correspondent()] = [
        Correspondent(),
        Fail(),
    ]
    # RequestService must receive back Ack or Fail
    P2PCommandAcks[RequestService()] = [
        Ack(),
        Fail(),
    ]
    # CancelService must receive back Ack or Fail
    P2PCommandAcks[CancelService()] = [
        Ack(),
        Fail(),
    ]
    P2PCommandAcks[Broadcast()] = []
    P2PCommandAcks[Relay()] = []
    P2PCommandAcks[RelayIn()] = []
    P2PCommandAcks[RelayOut()] = [
        RelayAck(),
        RelayFail(),
    ]
    P2PCommandAcks[RelayAck()] = []
    P2PCommandAcks[RelayFail()] = []
    P2PCommandAcks[Coin()] = [
        Ack(),
        Fail(),
    ]
    P2PCommandAcks[RetrieveCoin()] = [
        Coin(),
        Fail(),
    ]
    P2PCommandAcks[Key()] = [
        Ack(),
        Fail(),
    ]
    P2PCommandAcks[AuditKey()] = [
        Ack(),
        Fail(),
    ]
    P2PCommandAcks[Event()] = [
        Ack(),
        Fail(),
    ]
    P2PCommandAcks[Contacts()] = [
        Contacts(),
        Ack(),
        Fail(),
    ]
    # pre-define a set of commands for filtering out routed traffic
    RelayCommands = set([
        RelayIn(),
        RelayOut(),
        RelayAck(),
        RelayFail(),
        Relay(),
    ])


#------------------------------------------------------------------------------


def IsCommand(com):
    """
    Check to see if ``com`` is a valid command.
    """
    global P2PCommandAcks
    if P2PCommandAcks is None:
        init()
    return com in P2PCommandAcks


def IsCommandAck(com, ack):
    global P2PCommandAcks
    if P2PCommandAcks is None:
        init()
    return ack in P2PCommandAcks.get(com, [])


def IsAckExpected(com):
    global P2PCommandAcks
    if P2PCommandAcks is None:
        init()
    return Ack() in P2PCommandAcks.get(com, [])


def IsReplyExpected(com):
    global P2PCommandAcks
    if P2PCommandAcks is None:
        init()
    return len(P2PCommandAcks.get(com, [])) > 0


def IsRelay(com):
    global RelayCommands
    if RelayCommands is None:
        init()
    return com in RelayCommands


#------------------------------------------------------------------------------


def Ack():
    """
    Response packet for some request.
    """
    return 'Ack'


def Fail():
    """
    Used to report an error in response, for example when requested file is not
    found on remote machine.
    """
    return 'Fail'


def Identity():
    """
    Packet containing peer identity file.
    """
    return 'Identity'


def RequestService():
    return 'RequestService'


def CancelService():
    return 'CancelService'


def Key():
    return 'Key'


def AuditKey():
    return 'AuditKey'


def Event():
    return 'Event'


def Data():
    """
    Data packet, may be Data, Parity, Backup database, may be more.
    """
    return 'Data'


def Retrieve():
    """
    Used to request some data from supplier.
    """
    # TODO: rename to RetrieveData
    return 'Retrieve'


def Relay():
    """
    Used by proxy transport to route packets in/out via third node.
    """
    return 'Relay'


def RelayIn():
    """
    Used by proxy transport to route packets in/out via third node.
    """
    return 'RelayIn'


def RelayOut():
    """
    Used by proxy transport to route packets in/out via third node.
    """
    return 'RelayOut'


def RelayAck():
    """
    Used by proxy transport to route packets in/out via third node.
    """
    return 'RelayAck'


def RelayFail():
    """
    Used by proxy transport to route packets in/out via third node.
    """
    return 'RelayFail'


def ListFiles():
    """
    Response from remote peer with a list of my files stored on his machine.
    """
    return 'ListFiles'


def Files():
    """
    Request a list of my files from remote peer.
    """
    return 'Files'


def Contacts():
    """
    Packet with a list of my contacts, may be suppliers, customers or
    correspondents or list of suppliers of another customer.
    Can also be an empty packet with a request to provide some other contacts
    from remote peer.
    """
    return 'Contacts'


def DeleteFile():
    """
    Request to delete a single file or list of my files from remote machine.
    """
    return 'DeleteFile'


def DeleteBackup():
    """
    Request to delete whole backup or list of backups from remote machine.
    """
    return 'DeleteBackup'


def Transfer():
    """
    Transfer funds to remote peer.
    """
    # TODO: something for the future :)
    return 'Transfer'


def Receipt():
    """
    Some billing report.
    """
    # TODO: something for the future :)
    return 'Receipt'


def Message():
    """
    An encrypted message from one peer to another.
    Can be on of the types: "private_message", "queue_message", "group_message"
    """
    return 'Message'


def Correspondent():
    """
    Remote user should send you this to be included in your correspondents
    (friends) list.
    """
    return 'Correspondent'


def Broadcast():
    """
    This message type is for delivering some piece of data to all peers in th network.
    It is used to broadcast "crypto-coins" between peers.
    """
    # TODO: something for the future :)
    return 'Broadcast'


def Coin():
    # TODO: something for the future :)
    return 'Coin'


def RetrieveCoin():
    # TODO: something for the future :)
    return 'RetrieveCoin'



#!/usr/bin/python
# contract_chain_node.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (contract_chain_node.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred, DeferredList, succeed

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.p2p import commands
from bitdust.p2p import p2p_service

from bitdust.coins import coins_io
from bitdust.coins import contract_chain_consumer

#------------------------------------------------------------------------------


def reconnect():
    contract_chain_consumer.A('start')


def is_connected():
    if contract_chain_consumer.A() is None:
        lg.warn('contract_chain_consumer() not exist')
        return False
    if contract_chain_consumer.A().state != 'CONNECTED':
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_node._connected OFFLINE, nodes connection is not ready: %s' % contract_chain_consumer.A())
        return False
    return True


def get_coin_by_hash(hash_id):
    if not is_connected():
        return succeed(None)
    return Query(dict(
        method='get',
        index='hash',
        key=hash_id,
    ), ).result


def get_coins_by_chain(chain, provider_idurl, consumer_idurl):
    if not is_connected():
        return succeed(None)
    return Query(dict(
        method='get_many',
        index=chain,
        key='%s_%s' % (
            provider_idurl,
            consumer_idurl,
        ),
    ), ).result


def send_to_miner(coins):
    if not is_connected():
        return succeed(None)
    result = Deferred()
    p2p_service.SendCoin(
        contract_chain_consumer.A().connected_miner,
        coins,
        callbacks={
            commands.Ack(): lambda response, info: result.callback(response),
            commands.Fail(): lambda response, info: result.errback(Exception(response)),
        },
    )
    return result


#------------------------------------------------------------------------------


class Query(object):
    def __init__(self, query_dict):
        """
        """
        self.result = Deferred()
        self.out_packets = {}
        for idurl in contract_chain_consumer.A().connected_accountants:
            single_accountant = Deferred()
            outpacket = p2p_service.SendRetrieveCoin(
                idurl,
                query_dict,
                callbacks={
                    commands.Coin(): self._on_coin_received,
                    commands.Fail(): self._on_coin_failed,
                },
            )
            assert outpacket.PacketID not in self.out_packets
            self.out_packets[outpacket.PacketID] = single_accountant
        DeferredList(list(self.out_packets.values())).addBoth(self._on_results_collected)
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_node.Query created to request from %d accountants' % len(self.out_packets))

    def __del__(self):
        """
        """
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_node.Query object destoryed')

    def _close(self):
        """
        """
        self.result = None
        self.out_packets.clear()

    def _on_coin_received(self, response_packet, info):
        """
        """
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_node._on_coin_received %r' % response_packet)
        assert response_packet.PacketID in self.out_packets
        self.out_packets[response_packet.PacketID].callback((
            response_packet.CreatorID,
            response_packet,
        ))

    def _on_coin_failed(self, response_packet, info):
        """
        """
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_node._on_coin_failed %r' % response_packet)
        assert response_packet.PacketID in self.out_packets
        self.out_packets[response_packet.PacketID].callback((
            response_packet.CreatorID,
            None,
        ))

    def _on_results_collected(self, accountant_results):
        """
        """
        fails = {}
        unique_coins = {}
        for success, result_tuple in accountant_results:
            accountant_idurl, response_packet = result_tuple
            if not success:
                fails[accountant_idurl] = fails.get(accountant_idurl, 0) + 1
                continue
            if response_packet is None:
                fails[accountant_idurl] = fails.get(accountant_idurl, 0) + 1
                continue
            coins_list = coins_io.read_coins_from_packet(response_packet)
            if not coins_list:
                fails[accountant_idurl] = fails.get(accountant_idurl, 0) + 1
                continue
            for coin in coins_list:
                coin_hash = coin['miner']['hash']
                if coin_hash not in unique_coins:
                    unique_coins[coin_hash] = coin
                if coins_io.coin_to_string(unique_coins[coin_hash]) != coins_io.coin_to_string(coin):
                    fails[accountant_idurl] = fails.get(accountant_idurl, 0) + 1
        if not unique_coins:
            if _Debug:
                lg.out(_DebugLevel, 'contract_chain_node._on_results_collected : NO COINS FOUND')
            self.result.callback([])
            self._close()
            return None
        if fails:
            lg.warn('conflicting results from accountants: %s' % fails)
            # TODO: find some simple way to establish consensus between accountants
            # if one accountant is cheating or failing, coins from his machine must be filtered out
            # probably here we can trigger a process to replace bad accountants
            self.result.errback(Exception(str(fails)))
            self._close()
            return None
        if _Debug:
            lg.out(_DebugLevel, 'contract_chain_node._on_results_collected : %d COINS FOUND' % len(unique_coins))
        self.result.callback(list(unique_coins.values()))
        return None



#!/usr/bin/python
# settings.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (settings.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: settings.

I think this is a most used module in the project.
Various parts of the code use this module to get the user settings and global constants.

TODO:
need to move out user config stuff from that file
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

import os
import sys
import random

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.append(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..'))

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------
from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import deploy

from bitdust.lib import diskspace

from bitdust.main import config

#------------------------------------------------------------------------------

_UserConfig = None
_OverrideDict = {}  # list of specific values to be replacing some of the user settings
_InitDone = False

#------------------------------------------------------------------------------

_BandwidthLimit = None
_BackupBlockSize = None
_BackupMaxBlockSize = None
_P2PTimeOut = None

#------------------------------------------------------------------------------


def init(base_dir=None, network_name=None):
    """
    Must be called before all other things.
    This is called only once, prepare a bunch of things:

    - Set the base folder where for program data
    - Check and create if not exist "metadata" directory
    - Load settings from [app data dir]/[network name]/config or create a new, default settings
    - Check other "static" folders
    - Validate most important user settings
    - Check custom folders
    """
    global _InitDone
    if _InitDone:
        return
    _InitDone = True
    deploy.init_base_dir(base_dir=base_dir, network_name=network_name)
    if _Debug:
        lg.out(_DebugLevel, 'settings.init app data folder location is %r' % AppDataDir())
    _checkMetaDataDirectory()
    _checkConfigDirectory()
    _setUpDefaultSettings()
    _checkRandomizePortNumbers()
    _createNotExisingSettings()
    _checkStaticDirectories()
    _checkCustomDirectories()


def shutdown():
    global _InitDone
    _InitDone = False
    deploy.set_base_dir(None)
    config.shutdown()


#------------------------------------------------------------------------------
#---USER CONFIG----------------------------------------------------------------
#------------------------------------------------------------------------------


def override(key, value):
    """
    This method can be used to redefine values in UserConfig without writing
    changes on disk.

    Useful when user pass some params via command line - they should override the local settings.
    """
    global _OverrideDict
    if _Debug:
        lg.out(_DebugLevel, 'settings.override %s=%s' % (key, value))
    _OverrideDict[key] = value


def override_dict(d):
    """
    You can pass a dictionary of settings to override existing user config.
    """
    for key, value in d.items():
        override(key, value)


def convert_key(key):
    # try:
    key = key.replace('.', '/')
    key = key.replace('-enable', '-enabled')
    p = key.split('/')
    if len(p) == 1:
        return key
    if len(p) >= 3 and p[2].startswith(p[1] + '-'):
        p[2] = p[2].replace(p[1] + '-', '')
    if len(p) >= 2 and p[1].startswith(p[0] + '-'):
        p[1] = p[1].replace(p[0] + '-', '')
    if p[0] == 'folder':
        p[0] = 'paths'
    elif p[0] == 'backup':
        p[0] = 'services/backups'
        if p[1] == 'private-key-size':
            p[0] = 'personal'
    elif p[0] == 'general':
        p[0] = 'services/backups'
        if p[1] == 'backups':
            p[1] = 'max-copies'
        elif p[1] == 'local-backups-enabled':
            p[1] = 'keep-local-copies-enabled'
    elif p[0] == 'id-server':
        p[0] = 'services/identity-server'
    elif p[0] == 'identity-server':
        p[0] = 'services/identity-server'
    elif p[0] == 'network':
        p[0] = 'services/network'
        if p[1] == 'dht-port':
            p[0] = 'services/entangled-dht'
            p[1] = 'udp-port'
    elif p[0] == 'storage':
        if p[1] == 'donated':
            p[0] = 'services/supplier'
            p[1] = 'donated-space'
        elif p[1] == 'needed':
            p[0] = 'services/customer'
            p[1] = 'needed-space'
        elif p[1] == 'suppliers':
            p[0] = 'services/customer'
            p[1] = 'suppliers-number'
    elif p[0] == 'transport':
        if p[1] == 'tcp':
            p[0] = 'services'
            p[1] = 'tcp-transport'
            if len(p) > 2:
                if p[2] == 'port':
                    p[1] = 'tcp-connections'
                    p[2] = 'tcp-port'
        elif p[1] == 'udp':
            p[0] = 'services'
            p[1] = 'udp-transport'
            if len(p) > 2:
                if p[2] == 'port':
                    p[1] = 'udp-datagrams'
                    p[2] = 'udp-port'
    key = '/'.join(p)
    return key


"""
Below is a set of global constants.
"""

#------------------------------------------------------------------------------
#--- LOGS --------------------------------------------------------------------
#------------------------------------------------------------------------------


def UpdateLogFilename():
    """
    A place to store logs from update porcess.
    """
    return os.path.join(LogsDir(), 'software_update.log')


def AutomatsLog():
    """
    All state machines logs in the main process is written here.
    """
    return os.path.join(LogsDir(), 'automats.log')


def TransportLog():
    """
    Every x seconds will log stats about current transfers.
    """
    return os.path.join(LogsDir(), 'transport.log')


def ParallelPLogFilename():
    """
    Log from parallelp workers goes here, raid code is executed inside child processes.
    """
    return os.path.join(LogsDir(), 'parallelp.log')


def LocalTesterLogFilename():
    """
    A file name path where bptester.py will write its logs.
    """
    return os.path.join(LogsDir(), 'bptester.log')


#------------------------------------------------------------------------------
#--- CONSTANTS (NUMBERS) ------------------------------------------------------
#------------------------------------------------------------------------------


def MinimumIdentitySources():
    """
    You need to host your identity (piblic key, signature and contacts) at least in one place.
    By default you will use identity servers hard-coded in BitDust source code and maintained for you by enthusiasts.
    But you can also start your own identity server and host your identity on your own host.
    Use those settings before you run identity register call:

        services/identity-propagate/known-servers
        services/identity-propagate/preferred-servers
        services/identity-propagate/min-servers
        services/identity-propagate/max-servers
    """
    return 1


def MaximumIdentitySources():
    """
    You can host your identity in many places, up to 10 identity servers allowed.
    You do not need to mirror your identity too much, this is just to keep users
    consuming sufficient amount of network resources.
    """
    return 6


def DefaultPrivateKeySize():
    """
    User can choose size of his Private Key during install.

    Can be 1024, 2048 or 4096.
    """
    return 2048


def defaultDebugLevel():
    """
    Default debug level, lower values produce less messages.
    """
    return 0


def MaxPacketsOutstanding():
    """
    PREPRO Should be a function of disk space available.
    """
    return 100


def SendingSpeedLimit():
    """
    This is lower limit during file sending in bytes per second.

    Used to calculate a packet timeout - bigger packets should have longer timeouts.
    If sending below this speed - we count this supplier as failed.
    If we sending too slow to all nodes - it's our problems, not suppliers.
    """
    return 30*1024


def ReceivingSpeedLimit():
    return 30*1024


def DefaultBandwidthInLimit():
    """
    Incoming bandwidth limit in bytes per second, 0 - unlimited.
    """
    # 1 Mbps = 125000 B/s ~ 122 KB/s
    return 100*1250000


def DefaultBandwidthOutLimit():
    """
    Outgoing bandwidth limit in bytes per second, 0 - unlimited.
    """
    # 1 Mbps = 125000 B/s ~ 122 KB/s
    return 100*125000


def MaxRetries():
    """
    The idea was to create some "exponential backoff" -

    double each time. Set to 1 at the moment - so failed packets is ignored.
    """
    return 1


def DefaultEccMapName():
    """
    This is a ecc map name used by default.
    Must comply with ``DefaultDesiredSuppliers()``.
    """
    return 'ecc/2x2'


def DefaultDesiredSuppliers():
    """
    A starting number of suppliers for new users.

    Definitely we want to have 64 by default, but we need to have much
    more alive users to do that.

    Must comply with ``DefaultEccMapName()``
    """
    return 2


def MinimumUsernameLength():
    """
    A minimum possible user name length.
    """
    return 3


def MaximumUsernameLength():
    """
    A maximum possible user name length.
    """
    return 20


def DefaultDonatedBytes():
    """
    Default donated space value. User can set this at any moment in the settings.
    """
    return 8*1024*1024*1024  # 8 GB


def DefaultNeededBytes():
    """
    Default needed space value. User can set this at any moment in the settings.
    """
    return 256*1024*1024  # 256 MB


def MinimumDonatedBytes():
    """
    Minimum donated space amount in Megabytes - need to donate at least 2 Mb right now.
    """
    return 64*1024*1024  # 64 MB


def MinimumNeededBytes():
    return 32*1024*1024  # 32 MB - minimum 1 MB will be taken from every supplier


def DefaultBackupBlockSize():
    """
    Default block size in bytes, user can set this in settings. We split a
    backed up data into blocks of equal size and.

    perform RAID operation on every block - one by one.
    """
    return 4*1024*1024  # 4 MB


def DefaultBackupMaxBlockSize():
    """
    The maximum default block size, user can set this in the settings.
    """
    return 16*1024*1024  # 16 MB is fine


def MinimumBandwidthInLimitKBSec():
    """
    Not used, idea was to limit the minimum bandwidth given to BitDust.
    """
    return 10


def MinimumBandwidthOutLimitKBSec():
    """
    Not used.
    """
    return 10


def MaxDeletedBackupIDsToKeep():
    """
    How many deleted backup IDs do we want to hold on to in backup db.

    Not used.
    """
    return 100


#------------------------------------------------------------------------------
#--- CONSTANTS (TIMEOUTS/DELAYS) ------------------------------------------------------
#------------------------------------------------------------------------------


def P2PTimeOut():
    """
    A default timeout when sending and receiving packets.
    """
    global _P2PTimeOut
    if _P2PTimeOut is None:
        if config.conf():
            _P2PTimeOut = config.conf().getInt('services/gateway/p2p-timeout', 15)
        else:
            _P2PTimeOut = 15
    return _P2PTimeOut


def DefaultSendTimeOutEmail():
    """
    Timeout for email sending, not used.
    """
    return 300


def DefaultSendTimeOutHTTP():
    """
    Timeout for http sending, not used.
    """
    return 60


def DefaultAlivePacketTimeOut():
    """
    Lets send alive packets to our contacts every hour.
    """
    return 60*60


def DefaultBandwidthReportTimeOut():
    """
    Send ``BandwidthReport`` packets every 24 hours.
    """
    return 60*60*24


def DefaultNeedSuppliersPacketTimeOut():
    """
    If we need suppliers we will request it every 60 sec.

    Not used right now.
    """
    return 60


def DefaultLocaltesterLoop():
    """
    The code in ``p2p.local_tester`` will check the customers files
    periodically.

    This constant controls that - seconds between two calls.
    """
    return 20


def DefaultLocaltesterValidateTimeout():
    """
    A period in seconds to call ``Validate`` action of the local tester.
    """
    return 120*60


def DefaultLocaltesterUpdateCustomersTimeout():
    """
    A period in seconds to call ``UpdateCustomers`` action of the local tester.
    """
    return 5*60


def DefaultLocaltesterSpaceTimeTimeout():
    """
    A period in seconds to call ``SpaceTime`` action of the local tester.
    """
    return 5*60


def MinimumSendingDelay():
    """
    The lower limit of delay for repeated calls for sending processes.

    DO NOT SET TO 0 - the main process will be blocked.
    See ``lib.misc.LoopAttenuation`` method.
    """
    return 0.03


def MaximumSendingDelay():
    """
    The higher limit of delay for repeated calls for sending processes.

    Higher values should decrease the network speed, but save CPU
    resources.
    """
    return 4.0


def MinimumReceivingDelay():
    """
    Lower limit for receiving processes.
    """
    return 0.08


def MaximumReceivingDelay():
    """
    Higher limit for receiving processes.
    """
    return 4.0


def FireHireMinimumDelay():
    """
    Really do not want to fire suppliers too often, so use 15 minutes interval.
    """
    return 60*15


def BackupDBSynchronizeDelay():
    """
    Save backup index database no more than one time per every 5 min.
    """
    return 60*5


def SupplierContractBillingPeriodDays():
    """
    To prevent overhead on blockchain transactions, there is a limit.
    To pay for storage, you can send only one transaction to any user per billing period.
    So multiple contracts are paid at once.
    """
    return 30


#------------------------------------------------------------------------------
#---CONSTANTS ( STRINGS ) -----------------------------------------------------
#------------------------------------------------------------------------------


def ListFilesFormat():
    """
    Argument to ListFiles command to say format to return the data in.

    Can be "Text" or "Compressed". TODO: add "Encrypted" format
    """
    return 'Compressed'


def LegalUsernameChars():
    """
    A set of correct chars that can be used for user account names.
    """
    return set('abcdefghijklmnopqrstuvwxyz0123456789-_')


def LegalNickNameChars():
    """
    A set of correct chars that can be used for user account names.
    """
    return set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.,{}[]()@!$^&*=+')


#------------------------------------------------------------------------------
#--- FOLDERS ------------------------------------------------------------------
#------------------------------------------------------------------------------


def AppDataDir():
    """
    Return current data folder location, also call ``init()`` to be sure all
    things were configured.
    """
    init()
    return deploy.current_base_dir()


def BaseDir():
    return os.path.join(AppDataDir(), deploy.current_network())


def CurrentNetworkName():
    return deploy.current_network()


def BaseDirPathFileName():
    """
    You can configure BitDust software to use another place for data folder.

    Say you want to store BitDust files on another disk. In the binary
    folder file "appdata" can be created and it will keep the path to
    the data folder.
    """
    return os.path.join(bpio.getExecutableDir(), 'appdata')


def DefaultRestoreDir():
    """
    Default location to place restored files and folders.
    """
    if bpio.Android():
        # from android.storage import primary_external_storage_path  # @UnresolvedImport
        # return os.path.join(primary_external_storage_path(), 'Download')
        from android.storage import app_storage_path  # @UnresolvedImport
        return app_storage_path()
    return os.path.expanduser('~')


def MetaDataDir():
    """
    Return current location of the "metadata" folder - most important config files is here.
    """
    return os.path.join(BaseDir(), 'metadata')


def ConfigDir():
    return os.path.join(BaseDir(), 'config')


def TempDir():
    """
    A place for temporary BitDust files, we really need some extra disk space
    to operate.

    TODO: need to add some stuff to control how much extra space we use
    and be able limit that.
    """
    return os.path.join(AppDataDir(), 'temp')


def IdentityHistoryDir():
    """
    See ``lib.id_url`` module, this is a place to store all known idurl's to match
    "rotated" identies together.
    """
    return os.path.join(BaseDir(), 'identityhistory')


def IdentityCacheDir():
    """
    See ``lib.identitycache`` module, this is a place to store user's identity
    files to have them on hands.
    """
    return os.path.join(BaseDir(), 'identitycache')


def IdentityServerDir():
    return os.path.join(BaseDir(), 'identityserver')


def DefaultBackupsDBDir():
    """
    When you run the backup the following actions occur:

        - data is read from the local disk and compressed
        - entire volume is divided into blocks
        - blocks encrypted with user Key
        - each block is divided into pieces with redundancy - through RAID procedure
        - pieces of all blocks are stored on a local disk
        - pieces are transferred to suppliers
        - optionally, local pieces can be removed after delivering to suppliers

    This returns a default local folder location where those pieces is stored.
    User can configure that in the settings.
    """
    return os.path.join(BaseDir(), 'backups')


def DefaultCustomersDir():
    """
    Here will be placed files uploaded by other users.
    Your customers will user your HDD to keep their personal data.
    """
    return os.path.join(BaseDir(), 'customers')


def DefaultMessagesDir():
    """
    A default folder to store sent/received messages.
    """
    return os.path.join(BaseDir(), 'messages')


def DefaultMessagesHistoryDir():
    """
    A default folder to store sent/received messages.
    """
    return os.path.join(BaseDir(), 'messages', 'history')


def DefaultReceiptsDir():
    """
    A default folder to store receipts.
    """
    return os.path.join(BaseDir(), 'receipts')


def LogsDir():
    """
    Place for log files.
    """
    return os.path.join(AppDataDir(), 'logs')


def SuppliersDir():
    """
    Local folder location to keep suppliers info files.
    """
    return os.path.join(BaseDir(), 'suppliers')


def BandwidthInDir():
    """
    Daily stats for incoming bandwidth is placed in that location.
    """
    return os.path.join(BaseDir(), 'bandin')


def BandwidthOutDir():
    """
    Daily stats for outgoing bandwidth is placed in that location.
    """
    return os.path.join(BaseDir(), 'bandout')


def RatingsDir():
    """
    In that location BitDust software keeps a rating stats for known users.
    """
    return os.path.join(BaseDir(), 'ratings')


def ContractChainDir():
    return os.path.join(BaseDir(), 'contracts')


def ChatMessagesDir():
    return os.path.join(BaseDir(), 'messages')


def ChatChannelsDir():
    return os.path.join(BaseDir(), 'messages', 'channels')


def ChatHistoryDir():
    return os.path.join(BaseDir(), 'messages', 'history')


def KeyStoreDir():
    return os.path.join(BaseDir(), 'keys')


def BlockchainDir():
    return os.path.join(BaseDir(), 'blockchain')


def APICertificatesDir():
    return os.path.join(BaseDir(), 'apicerts')


def ServicesDataDir():
    return os.path.join(BaseDir(), 'servicedata')


def ServiceDir(service_name):
    return os.path.join(ServicesDataDir(), service_name.replace('service_', ''))


#------------------------------------------------------------------------------
#--- FILES --------------------------------------------------------------------
#------------------------------------------------------------------------------


def KeyFileName():
    """
    Location of user's Private Key file.
    """
    return os.path.join(MetaDataDir(), 'mykeyfile')


def KeyFileNameLocation():
    """
    User can set another location for his Private Key file - he can use USB stick to keep his Key.
    After BitDust stars he can remove the USB stick and keep it in safe place.
    So BitDust will keep user's key in the RAM only - this way you can have more protection for your Key.
    If your machine is stolen - thief can not get your Private key.
    But you must be sure that machine was switched off - the RAM is erased when power is off.
    This file keeps alternative location of your Private Key.
    """
    return KeyFileName() + '_location'


def CustomerIDsFilename():
    """
    IDs for places we store data for, keeps a list of IDURLs of our customers.
    """
    return os.path.join(MetaDataDir(), 'customerids')


def CustomersMetaInfoFilename():
    """
    Keeps a list of known ECCMaps and other info of my customers.
    """
    return os.path.join(MetaDataDir(), 'customersmetainfo')


def CorrespondentIDsFilename():
    """
    People we get messages from and other stuff not related to backup/restore
    process.
    """
    return os.path.join(MetaDataDir(), 'correspondentids')


def LocalIdentityFilename():
    """
    A local copy of user's identity file is stored here.

    When doing any changes in the identity file - this appear here firstly.
    Further local identity file is propagated to the identity server
    and all our contacts so they got the fresh copy asap.
    """
    return os.path.join(MetaDataDir(), 'localidentity')


def LocalIPFilename():
    """
    File contains string like "192.168.12.34" - local IP of that machine.
    """
    return os.path.join(MetaDataDir(), 'localip')


def ExternalIPFilename():
    """
    File contains string like 201.42.133.2 - external IP of that machine.
    """
    return os.path.join(MetaDataDir(), 'externalip')


def ExternalUDPPortFilename():
    """
    File contains external UDP port number of that machine - detected after STUN.
    """
    return os.path.join(MetaDataDir(), 'externaludpport')


def DefaultTransportOrderFilename():
    """
    Location for file that keeps an order of used transports.
    """
    return os.path.join(MetaDataDir(), 'torder')


def UserNameFilename():
    """
    File contains something like "guesthouse" - user account name.
    """
    return os.path.join(MetaDataDir(), 'username')


def BackupIndexFileName():
    """
    This is backup data base index file name.
    The file stores folder and files names and locations with path ID's and some extra info.
    Also this file is saved on suppliers in encrypted state.
    """
    return 'index'


def BackupIndexFilePath(customer_idurl=None, key_alias='master'):
    from bitdust.userid import my_id
    from bitdust.userid import id_url
    if customer_idurl is None:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    customer_id = customer_idurl.to_id()
    index_dir_path = os.path.join(ServiceDir('service_backups'), 'index')
    index_file_path = os.path.join(index_dir_path, '%s$%s' % (key_alias, customer_id))
    return index_file_path


def SupplierPath(supplier_idurl, customer_idurl, filename=None):
    """
    A location to given supplier's data.

    If ``filename`` is provided - return a full path to that file.
    Currently those data are stored for every supplier:

        - "connected" : date and time when this man become our suppler
        - "disconnected" : date and time when this suppler was fired
        - "listfiles" : a list of our local files stored on his machine
    """
    from bitdust.userid import global_id
    from bitdust.lib import nameurl
    customer = global_id.UrlToGlobalID(customer_idurl)
    if filename is not None:
        return os.path.join(SuppliersDir(), customer, nameurl.UrlFilename(supplier_idurl), filename)
    return os.path.join(SuppliersDir(), customer, nameurl.UrlFilename(supplier_idurl))


def SupplierListFilesFilename(supplier_idurl, customer_idurl):
    """
    Return a "listfiles" file location for given supplier.
    """
    return os.path.join(SupplierPath(supplier_idurl, customer_idurl), 'listfiles')


def SupplierServiceFilename(supplier_idurl, customer_idurl):
    """
    Return a "service" file location for given supplier.
    """
    return os.path.join(SupplierPath(supplier_idurl, customer_idurl), 'service')


def RepoFile():
    """
    A file to store info about currently used repository.
    """
    return os.path.join(MetaDataDir(), 'repo')


def VersionNumberFile():
    return os.path.join(bpio.getExecutableDir(), 'version')


def CheckSumFile():
    return os.path.join(bpio.getExecutableDir(), 'checksum')


def CustomersSpaceFile():
    """
    This file keeps info about our customers - how many megabytes every guy takes from us.
    """
    return os.path.join(MetaDataDir(), 'space')


def CustomersUsedSpaceFile():
    return os.path.join(MetaDataDir(), 'spaceused')


def BalanceFile():
    """
    This file keeps our current BitDust balance - two values:
        - transferable funds
        - not transferable funds
    """
    return os.path.join(MetaDataDir(), 'balance')


def CertificateFiles():
    """
    The idea is to have a global certificate for BitDust server, just like
    https works.
    """
    return [
        os.path.join(MetaDataDir(), 'bitdust.cer'),
        os.path.join('.', 'bitdust.cer'),
        os.path.join(bpio.getExecutableDir(), 'bitdust.cer'),
    ]


def DHTDBFile():
    return os.path.join(MetaDataDir(), 'dhtdb')


def FTPServerCredentialsFile():
    return os.path.join(MetaDataDir(), 'ftpcredentials')


def APIServerCertificateKeyFile():
    return os.path.join(AppDataDir(), 'apiservercertkey')


def APIServerCertificateFile():
    return os.path.join(AppDataDir(), 'apiservercert')


def APIClientCertificateKeyFile():
    return os.path.join(AppDataDir(), 'apiclientcertkey')


def APIClientCertificateFile():
    return os.path.join(AppDataDir(), 'apiclientcert')


def APISecretFile():
    return os.path.join(AppDataDir(), 'apisecret')


def ChatMessagesHistoryDatabaseFile():
    return os.path.join(ChatMessagesDir(), 'message_v1.db')


#------------------------------------------------------------------------------
#--- BINARY FILES -------------------------------------------------------------
#------------------------------------------------------------------------------


def WindowsStarterFileName():
    """
    Return a file name of the Windows starter: "bitstarter.exe".
    Obsolete.
    """
    return 'bitstarter.exe'


def WindowsMainScriptFileName():
    """
    Return a file name of the root Windows executable script : "bitdust-node.exe".
    """
    return 'bitdust-node.exe'


def getIconLaunchFilename():
    """
    Not used.

    For Windows platforms this should target to executable file to run
    when clicked on Desktop icon.
    """
    return os.path.join(bpio.getExecutableDir(), 'bitdust-node.exe')


def IconsFolderPath():
    """
    A folder name where application icons is stored.

    PREPRO: maybe we better use another name: "media", because we may
    need not only "icons" but also other data files
    """
    return os.path.join(bpio.getExecutableDir(), 'icons')


def FontsFolderPath():
    """
    A folder name where application "fons" is stored.
    """
    return os.path.join(bpio.getExecutableDir(), 'fonts')


def FontImageFile():
    """
    A font to use to print text labels in the GUI.
    """
    return os.path.join(FontsFolderPath(), 'Arial_Narrow.ttf')


#------------------------------------------------------------------------------
#---PORT NUMBERS---------------------------------------------------------------
#------------------------------------------------------------------------------


def DefaultXMLRPCPort():
    """
    Only Local! Never expose to outside of localhost.
    Obsolete. To be removed.
    """
    return 8082


def DefaultJsonRPCPort():
    """
    Only Local! Never expose to outside of localhost.
    """
    return 8083


def DefaultRESTHTTPPort():
    """
    Only Local! Never expose to outside of localhost.
    """
    return 8180


def DefaultWebSocketPort():
    """
    Only Local! Never expose to outside of localhost.
    """
    return 8280


def DefaultFTPPort():
    """
    Only Local! Never expose to outside of localhost.
    """
    return 8021


def IdentityServerPort():
    """
    Identity server stores identity files, it works in that way:

        1) anyone can request any stored identity file from any place in the world
        2) anyone can send his identity file over transport_tcp to identity server
        3) identity file must be digitaly signed, server should verify the signature
        4) if signature is fine - server will save (or overwrite existing) the file
        5) server should refuse incorrect or faked identities
        6) someone can store incorrect or faked identities on his own server, but nodes in network will refuse those identities
        7) you can use different ways to transfer your identity file to your own id server - do it by your own.

    This is a port number of our identity file to receive identity files from users.
    Must accept external incoming connections.
    """
    return 6661  # 7773


def IdentityWebPort():
    """
    Our public identity servers use standard web port number to publish
    identity files.
    Must accept external incoming connections.
    Normally you would add port 8084 to port 80 routing using Apache2 or Nginx web server.
    """
    return 8084


def DefaultTCPPort():
    """
    A default port number for transport_tcp.
    Must accept external incoming connections.
    """
    return 7771


def DefaultUDPPort():
    """
    A default port number for transport_udp.
    Must accept external incoming connections.
    """
    return 8882


def DefaultDHTPort():
    """
    A default UDP port number for DHT network.
    Must accept external incoming connections.
    """
    return 14441


def DefaultHTTPPort():
    """
    A default port number for transport_http.
    Must accept external incoming connections.
    """
    return 9993


def DefaultWebLogPort():
    """
    A port number for HTTP server to print program logs.
    Only Local! Never expose to outside of localhost.
    """
    return 9999


def DefaultWebTrafficPort():
    """
    A port number for HTTP server to print program packets traffic.
    Only Local! Never expose to outside of localhost.
    """
    return 9997


#------------------------------------------------------------------------------
#--- USER FOLDERS -------------------------------------------------------------
#------------------------------------------------------------------------------


def getCustomersFilesDir():
    """
    Alias to get a user donated location from settings.
    """
    return config.conf().getString('paths/customers', default=DefaultCustomersDir()).strip()


def getCustomerFilesDir(idurl):
    """
    Alias to get a given customer's files inside our donated location from
    settings.
    """
    from bitdust.userid import global_id
    return os.path.join(getCustomersFilesDir(), global_id.UrlToGlobalID(idurl))


def getLocalBackupsDir():
    """
    Alias to get local backups folder from settings, see
    ``DefaultBackupsDBDir()``.
    """
    return config.conf().getString('paths/backups', default=DefaultBackupsDBDir()).strip()


def getRestoreDir():
    """
    Alias for restore location, see ``DefaultRestoreDir()``.
    """
    return config.conf().getString('paths/restore', default=DefaultRestoreDir()).strip()


def getReceiptsDir():
    """
    Alias to get from user config a folder location where receipts is stored.
    """
    return config.conf().getString('paths/receipts', default=DefaultReceiptsDir()).strip()


def getTempDir():
    """
    An alias for ``TempDir()``.
    """
    return TempDir()


#------------------------------------------------------------------------------
#--- OS PROXY SERVER OPTIONS --------------------------------------------------
#------------------------------------------------------------------------------


def enableLocalProxy(enable=None):
    """
    Enable/disable using of proxy server.
    """
    if enable is None:
        return config.conf().getBool('services/network/proxy/enabled')
    config.conf().setData('services/network/proxy/enabled', str(enable))


def getProxyHost():
    """
    Return proxy server host from settings.
    """
    return config.conf().getData('services/network/proxy/host').strip()


def getProxyPort():
    """
    Return proxy server port number from settings.
    """
    return config.conf().getData('services/network/proxy/port').strip()


def setProxySettings(d):
    """
    Set proxy settings via dictionary, see
    ``lib.net_misc.detect_proxy_settings`` for more details.
    """
    if 'host' in d:
        config.conf().setData('services/network/proxy/host', str(d.get('host', '')))
    if 'port' in d:
        config.conf().setData('services/network/proxy/port', str(d.get('port', '')))
    if 'username' in d:
        config.conf().setData('services/network/proxy.network-proxy-username', str(d.get('username', '')))
    if 'password' in d:
        config.conf().setData('services/network/proxy/password', str(d.get('password', '')))
    if 'ssl' in d:
        config.conf().setData('services/network/proxy/ssl', str(d.get('ssl', 'False')))


def getProxySettingsDict():
    """
    Return a proxy settings from user config in dictionary.
    """
    return {
        'host': config.conf().getData('services/network/proxy/host').strip(),
        'port': config.conf().getData('services/network/proxy/port').strip(),
        'username': config.conf().getData('services/network/proxy/username').strip(),
        'password': config.conf().getData('services/network/proxy/password').strip(),
        'ssl': config.conf().getData('services/network/proxy/ssl').strip(),
    }


def update_proxy_settings():
    """
    Calls ``lib.net_misc.detect_proxy_settings()`` to check current system
    proxy server settings.
    """
    from bitdust.lib import net_misc
    net_misc.init()
    if enableLocalProxy():
        if getProxyHost() == '' or getProxyPort() == '':
            d = net_misc.detect_proxy_settings()
            # setProxySettings(d)
            # enableLocalProxy(d.get('host', '') != '')
            if _Debug:
                lg.out(_DebugLevel, 'settings.update_proxy_settings : loaded from ENVIRONMENT')
        else:
            d = getProxySettingsDict()
            if _Debug:
                lg.out(_DebugLevel, 'settings.update_proxy_settings : loaded from settings')
        net_misc.set_proxy_settings(d)
        if _Debug:
            lg.out(_DebugLevel, '    HOST:      ' + net_misc.get_proxy_host())
            lg.out(_DebugLevel, '    PORT:      ' + str(net_misc.get_proxy_port()))
            lg.out(_DebugLevel, '    USERNAME:  ' + net_misc.get_proxy_username())
            lg.out(_DebugLevel, '    PASSWORD:  ' + ('*'*len(net_misc.get_proxy_password())))
            lg.out(_DebugLevel, '    SSL:       ' + net_misc.get_proxy_ssl())


#------------------------------------------------------------------------------
#---OTHER USER CONFIGURATIONS--------------------------------------------------
#------------------------------------------------------------------------------


def getBandOutLimit():
    """
    This is a current outgoing bandwidth limit in bytes per second.
    """
    return config.conf().getInt('services/network/send-limit', DefaultBandwidthOutLimit())


def getBandInLimit():
    """
    This is a current incoming bandwidth limit in bytes per second.
    """
    return config.conf().getInt('services/network/receive-limit', DefaultBandwidthInLimit())


def enableIdServer(enable=None):
    if enable is None:
        return config.conf().getBool('services/identity-server/enabled')
    config.conf().setData('services/identity-server/enabled', str(enable))


def enableAPIAuthSecret(enable=None):
    if enable is None:
        return config.conf().getBool('interface/api/auth-secret-enabled')
    config.conf().setData('interface/api/auth-secret-enabled', str(enable))


def enableRESTHTTPServer(enable=None):
    if enable is None:
        return config.conf().getBool('interface/api/rest-http-enabled')
    config.conf().setData('interface/api/rest-http-enabled', str(enable))


def enableWebSocketServer(enable=None):
    if enable is None:
        return config.conf().getBool('interface/api/web-socket-enabled')
    config.conf().setData('interface/api/web-socket-enabled', str(enable))


def enableFTPServer(enable=None):
    if enable is None:
        return config.conf().getBool('interface/ftp/enabled')
    config.conf().setData('interface/ftp/enabled', str(enable))


def getIdServerHost():
    return config.conf().getString('services/identity-server/host').strip()


def setIdServerHost(hostname_or_ip):
    return config.conf().setString('services/identity-server/host', hostname_or_ip)


def getIdServerWebPort():
    return config.conf().getInt('services/identity-server/web-port', IdentityWebPort())


def setIdServerWebPort(web_port):
    return config.conf().setInt('services/identity-server/web-port', web_port)


def getIdServerTCPPort():
    return config.conf().getInt('services/identity-server/tcp-port', IdentityServerPort())


def setIdServerTCPPort(tcp_port):
    return config.conf().setInt('services/identity-server/tcp-port', tcp_port)


def getRESTHTTPServerPort():
    return config.conf().getInt('interface/api/rest-http-port', DefaultRESTHTTPPort())


def setRESTHTTPServerPort(rest_http_port):
    return config.conf().setInt('interface/api/rest-http-port', rest_http_port)


def getWebSocketServerPort():
    return config.conf().getInt('interface/api/web-socket-port', DefaultWebSocketPort())


def setWebSocketServerPort(rest_http_port):
    return config.conf().setInt('interface/api/web-socket-port', rest_http_port)


def getFTPServerPort():
    return config.conf().getInt('interface/ftp/port', DefaultFTPPort())


def setFTPServerPort(ftp_port):
    return config.conf().setInt('interface/ftp/port', ftp_port)


def getTransportPort(proto):
    """
    Get a port number for some tranports from user config.
    """
    if proto == 'tcp':
        return getTCPPort()
    if proto == 'udp':
        return getUDPPort()
    if proto == 'http':
        return getHTTPPort()
    raise


def enableTCP(enable=None):
    """
    Switch on/off transport_tcp in the settings or get current state.

    Note : transport_tcp is always available for identites to id server.
    """
    if enable is None:
        return config.conf().getBool('services/tcp-transport/enabled')
    config.conf().setData('services/tcp-transport/enabled', str(enable))


def enableTCPsending(enable=None):
    """
    Switch on/off sending over transport_tcp in the settings or get current
    state.
    """
    if enable is None:
        return config.conf().getBool('services/tcp-transport/sending-enabled')
    config.conf().setData('services/tcp-transport/sending-enabled', str(enable))


def enableTCPreceiving(enable=None):
    """
    Switch on/off receiving over transport_tcp in the settings or get current
    state.
    """
    if enable is None:
        return config.conf().getBool('services/tcp-transport/receiving-enabled')
    config.conf().setData('services/tcp-transport/receiving-enabled', str(enable))


def getTCPPort():
    """
    Get a port number for tranport_tcp from user config.
    """
    return config.conf().getInt('services/tcp-connections/tcp-port', DefaultTCPPort())


def setTCPPort(port):
    """
    Set a port number for tranport_tcp in the user config.
    """
    config.conf().setData('services/tcp-connections/tcp-port', str(port))


def enableUDP(enable=None):
    """
    Switch on/off transport_udp in the settings or get current state.
    """
    if enable is None:
        return config.conf().getBool('services/udp-transport/enabled')
    config.conf().setData('services/udp-transport/enabled', str(enable))


def enableUDPsending(enable=None):
    """
    Switch on/off sending over udp in the settings or get current state.
    """
    if enable is None:
        return config.conf().getBool('services/udp-transport/sending-enabled')
    config.conf().setData('services/udp-transport/sending-enabled', str(enable))


def enableUDPreceiving(enable=None):
    """
    Switch on/off receiving over udp in the settings or get current state.
    """
    if enable is None:
        return config.conf().getBool('services/udp-transport/receiving-enabled')
    config.conf().setData('services/udp-transport/receiving-enabled', str(enable))


def getUDPPort():
    """
    Get a port number for tranport_udp from user config.
    """
    return config.conf().getInt('services/udp-datagrams/udp-port', DefaultUDPPort())


def setUDPPort(port):
    """
    Set a port number for tranport_udp in the user config.
    """
    config.conf().setData('services/udp-datagrams/udp-port', str(port))


def getDHTPort():
    """
    Get a UDP port number for entangled "DHT" network.
    """
    return config.conf().getInt('services/entangled-dht/udp-port', DefaultDHTPort())


def enablePROXY(enable=None):
    """
    Switch on/off transport_proxy in the settings or get its current state.
    """
    if enable is None:
        return config.conf().getBool('services/proxy-transport/enabled')
    config.conf().setData('services/proxy-transport/enabled', str(enable))


def enablePROXYsending(enable=None):
    """
    Switch on/off sending over proxy transport in the settings or get current
    state.
    """
    if enable is None:
        return config.conf().getBool('services/proxy-transport/sending-enabled')
    config.conf().setData('services/proxy-transport/sending-enabled', str(enable))


def enablePROXYreceiving(enable=None):
    """
    Switch on/off receiving over proxy transport in the settings or get current
    state.
    """
    if enable is None:
        return config.conf().getBool('services/proxy-transport/receiving-enabled')
    config.conf().setData('services/proxy-transport/receiving-enabled', str(enable))


def enableHTTP(enable=None):
    """
    Switch on/off transport_http in the settings or get current state.
    """
    if enable is None:
        return config.conf().getBool('services/http-transport/enabled')
    config.conf().setData('services/http-transport/enabled', str(enable))


def enableHTTPsending(enable=None):
    """
    Switch on/off sending over transport_http in the settings or get current
    state.
    """
    if enable is None:
        return config.conf().getBool('services/http-transport/sending-enabled')
    config.conf().setData('services/http-transport/sending-enabled', str(enable))


def enableHTTPreceiving(enable=None):
    """
    Switch on/off receiving over transport_http in the settings or get current
    state.
    """
    if enable is None:
        return config.conf().getBool('services/http-transport/receiving-enabled')
    config.conf().setData('services/http-transport/receiving-enabled', str(enable))


def getHTTPPort():
    """
    Get a port number for tranport_http from user config.
    """
    return config.conf().getInt('services/http-connections/http-port', DefaultTCPPort())


def setHTTPPort(port):
    """
    Set a port number for tranport_http in the user config.
    """
    config.conf().setData('services/http-connections/http-port', str(port))


def getTransportPriority(proto):
    return config.conf().getInt('services/%s-transport/priority' % proto, 1)


def setTransportPriority(proto, value):
    return config.conf().setInt('services/%s-transport/priority' % proto, value)


def setDHTPort(port):
    """
    Set a UDP port number for entangled "DHT" network.
    """
    config.conf().setData('services/entangled-dht/udp-port', str(port))


def enableTransport(proto, enable=None):
    """
    Return a current state of given network transport or set a new state.
    """
    # key = 'transport.transport-%s.transport-%s-enable' % (proto, proto)
    key = 'services/%s-transport/enabled' % proto
    if config.conf().getData(key) is None:
        return False
    if enable is None:
        return config.conf().getBool(key)
    config.conf().setData(key, str(enable))


def transportIsEnabled(proto):
    """
    Alias for ``enableTransport()``.
    """
    return enableTransport(proto)


def transportIsInstalled(proto):
    """
    This should return True if given transport have been configured and all
    needed config info is available.
    """
    return True


def transportReceivingIsEnabled(proto):
    """
    Return True if receiving over given transport is switched on.
    """
    # key = 'transport.transport-%s.transport-%s-receiving-enable' % (proto, proto)
    key = 'services/%s-transport/receiving-enabled' % proto
    if config.conf().getData(key) is None:
        return False
    return config.conf().getBool(key)


def transportSendingIsEnabled(proto):
    """
    Return True if sending over given transport is switched on.
    """
    # key = 'transport.transport-%s.transport-%s-sending-enable' % (proto, proto)
    key = 'services/%s-transport/sending-enabled' % proto
    if config.conf().getData(key) is None:
        return False
    return config.conf().getBool(key)


def enableProxyServer(enable=None):
    if enable is None:
        return config.conf().getBool('services/proxy-server/enabled')
    config.conf().setData('services/proxy-server/enabled', str(enable))


def getDebugLevelStr():
    """
    This is just for checking if it is set, the int() would throw an error.
    """
    return config.conf().getData('logs/debug-level')


def getDebugLevel():
    """
    Return current debug level.
    """
    return config.conf().getInt('logs/debug-level', lg._GlobalDebugLevel)


def setDebugLevel(level):
    """
    Set debug level.
    """
    config.conf().setData('logs/debug-level', str(level))


def enableWebStream(enable=None):
    """
    Get current state or enable/disable using of HTTP server to print logs,
    need to restart BitDust to take place changes.
    """
    if enable is None:
        return config.conf().getBool('logs/stream-enabled')
    config.conf().setData('logs/stream-enabled', str(enable))


def enableWebTraffic(enable=None):
    """
    Get current state or enable/disable using of HTTP server to print packets
    traffic, need to restart BitDust to take place changes.
    """
    if enable is None:
        return config.conf().getBool('logs/traffic-enabled')
    config.conf().setData('logs/traffic-enabled', str(enable))


def getWebStreamPort():
    """
    Get port number of HTTP server to print logs.
    """
    return config.conf().getInt('logs/stream-port', DefaultWebLogPort())


def getWebTrafficPort():
    """
    Get port number of HTTP server to print packets traffic.
    """
    return config.conf().getInt('logs/traffic-port')


def enableMemoryProfile(enable=None):
    """
    Get current state or enable/disable using of HTTP server to momory
    profiling.
    """
    if enable is None:
        return config.conf().getBool('logs/memprofile-enabled')
    return config.conf().setData('logs/memprofile-enabled', str(enable))


def getECCSuppliersNumbers():
    """
    List of available suppliers numbers.
    """
    return [2, 4, 7, 13]
    # return eccmap.SuppliersNumbers()


def getSuppliersNumberDesired():
    """
    Get suppliers number from user settings.
    """
    return config.conf().getInt('services/customer/suppliers-number', -1)


def getNeededString():
    """
    Get needed space in megabytes from user settings.
    """
    return config.conf().getData('services/customer/needed-space')


def getNeededBytes():
    return diskspace.GetBytesFromString(getNeededString())


def getDonatedString():
    """
    Get donated space in megabytes from user settings.
    """
    return config.conf().getData('services/supplier/donated-space')


def getDonatedBytes():
    return diskspace.GetBytesFromString(getDonatedString())


def getUpdatesMode():
    """
    User can set different modes to update the BitDust software.
    """
    # TODO: remove this after get rid of webcontrol.py
    return 'turn off updates'


def getUpdatesModeValues():
    """
    List of available update modes.
    """
    # TODO: remove this after get rid of webcontrol.py
    return (
        'install automatically',
        'only notify',
        'turn off updates',
    )


def getUpdatesSheduleData():
    """
    Return update schedule from settings.
    """
    return config.conf().getData('updates/shedule')


def setUpdatesSheduleData(raw_shedule):
    """
    Set update schedule in the settings.
    """
    config.conf().setData('updates/shedule', raw_shedule)


def getBackupsMaxCopies():
    """
    Return a number of copies to keep for every backed up data.

    The oldest copies (over that amount) will be removed from data base
    and remote suppliers.
    """
    return config.conf().getInt('services/backups/max-copies', 2)


def getBackupsKeepLocalCopies():
    """
    Return True if user wish to keep local backups.
    """
    return config.conf().getBool('services/backups/keep-local-copies-enabled')


def getGeneralWaitSuppliers():
    """
    Return True if user want to be sure that suppliers are reliable enough
    before removing the local backups.
    """
    return config.conf().getBool('services/backups/wait-suppliers-enabled')


def getBackupBlockSizeStr():
    return config.conf().getData('services/backups/block-size')


def getBackupBlockSize():
    """
    Get backup block size from settings.
    """
    return diskspace.GetBytesFromString(getBackupBlockSizeStr())


def getBackupMaxBlockSizeStr():
    return config.conf().getData('services/backups/max-block-size')


def getBackupMaxBlockSize():
    """
    Get the maximum backup block size from settings.
    """
    return diskspace.GetBytesFromString(getBackupMaxBlockSizeStr())


def setBackupBlockSize(block_size):
    """
    Set current backup block size in the memory to have fast access.
    """
    return config.conf().setData('services/backups/block-size', diskspace.MakeStringFromBytes(block_size))


def setBackupMaxBlockSize(block_size):
    """
    Set current maximum backup block size in the memory to have fast access.
    """
    return config.conf().setData('services/backups/max-block-size', diskspace.MakeStringFromBytes(block_size))


def getPrivateKeySize():
    """
    Returns your preferred Private Key size from settings.
    Your Master Private Key is generated only once during install stage.
    Other Private Keys also will be generated when necessary:
    for customer-supplier interactions, for private messages, for blockchain service, etc.
    """
    return config.conf().getInt('personal/private-key-size')


def setPrivateKeySize(pksize):
    """
    Set Private Key size in the settings.
    """
    config.conf().setInt('personal/private-key-size', pksize)


def enableUPNP(enable=None):
    """
    Return True if user want to try to config his router to config port
    forwarding automatically.

    If ``enable`` is not None - rewrite current value in the settings.
    """
    if enable is None:
        return config.conf().getBool('services/tcp-connections/upnp-enabled')
    config.conf().setData('services/tcp-connections/upnp-enabled', str(enable))


def enableService(svc_name, enable=None):
    if enable is None:
        return config.conf().getBool('services/%s/enabled' % svc_name)
    config.conf().setBool('services/%s/enabled' % svc_name, enable)


def enableBroadcastRouting(enable=None):
    if enable is None:
        return config.conf().getBool('services/broadcasting/routing-enabled')
    config.conf().setBool('services/broadcasting/routing-enabled', enable)


#------------------------------------------------------------------------------
#--- USER SETTINGS VALIDATION -------------------------------------------------
#------------------------------------------------------------------------------


def _checkMetaDataDirectory():
    """
    Check that the __metadata__ directory exists.
    """
    if os.path.isfile(MetaDataDir()):
        raise Exception('file already exist:' + MetaDataDir())
    if not os.path.exists(MetaDataDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings._checkMetaDataDirectory want to create "metadata" folder in : ' + MetaDataDir())
        bpio._dirs_make(MetaDataDir())
        return
    if _Debug:
        lg.out(_DebugLevel, 'settings._checkMetaDataDirectory OK , folder already exist: ' + MetaDataDir())


def _checkConfigDirectory():
    """
    Check that the __config__ directory exists.
    """
    if os.path.isfile(ConfigDir()):
        raise Exception('file already exist:' + ConfigDir())
    if not os.path.exists(ConfigDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings._checkConfigDirectory want to create "config" folder in : ' + ConfigDir())
        bpio._dir_make(ConfigDir())
    else:
        if _Debug:
            lg.out(_DebugLevel, 'settings._checkConfigDirectory OK , folder already exist: ' + ConfigDir())
    config.init(ConfigDir())


def _setUpDefaultSettings():
    """
    Configure default values for all settings.

    Every option must have a default value, howerver there are exceptions possible :-)
    """
    from bitdust.main import config_defaults
    config_defaults.reset(config.conf())


def _checkRandomizePortNumbers():
    """
    To avoid conflicts between two nodes inside same sub-network they both need to use
    different port numbers. So this method will first check if port number already set or not.
    If it is not it will set random value in the range.
    """
    # 7000-8000 for tcp transport
    if not config.conf().getOriginalData('services/tcp-connections/tcp-port'):
        config.conf().setData('services/tcp-connections/tcp-port', str(random.randint(7001, 8000) - 1))
    # 8000-9000 for udp transport
    if not config.conf().getOriginalData('services/udp-datagrams/udp-port'):
        config.conf().setData('services/udp-datagrams/udp-port', str(random.randint(8001, 9000) - 1))
    # 9000-10000 for http transport
    if not config.conf().getOriginalData('services/http-connections/http-port'):
        config.conf().setData('services/http-connections/http-port', str(random.randint(9001, 10000) - 1))
    # 10000-11000 for entangled dht
    if not config.conf().getOriginalData('services/entangled-dht/udp-port'):
        config.conf().setData('services/entangled-dht/udp-port', str(random.randint(10001, 11000) - 1))
    # 11000-12000 for blockchain
    if not config.conf().getOriginalData('services/blockchain/port'):
        config.conf().setData('services/blockchain/port', str(random.randint(11001, 12000) - 1))


def _createNotExisingSettings():
    """
    Validate user settings and create them from default values.
    """
    for key in config.conf()._default.keys():
        if not config.conf().exist(key):
            value = config.conf().getDefaultValue(key)
            config.conf().setData(key, value)
            if _Debug:
                lg.out(_DebugLevel, '    created option %s with default value : [%s]' % (key, value))
            # print '    created option %s with default value : [%s]' % (key, value)


def _checkStaticDirectories():
    """
    Check existance of static data folders.
    """
    if not os.path.exists(BaseDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + BaseDir())
        os.makedirs(BaseDir())
    if not os.path.exists(TempDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + TempDir())
        os.makedirs(TempDir())
#     if not os.path.exists(BandwidthInDir()):
#         if _Debug:
#             lg.out(_DebugLevel, 'settings.init want to create folder: ' + BandwidthInDir())
#         os.makedirs(BandwidthInDir())
#     if not os.path.exists(BandwidthOutDir()):
#         if _Debug:
#             lg.out(_DebugLevel, 'settings.init want to create folder: ' + BandwidthOutDir())
#         os.makedirs(BandwidthOutDir())
    if not os.path.exists(LogsDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + LogsDir())
        os.makedirs(LogsDir())
    if not os.path.exists(IdentityCacheDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + IdentityCacheDir())
        os.makedirs(IdentityCacheDir())
    if not os.path.exists(SuppliersDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + SuppliersDir())
        os.makedirs(SuppliersDir())
    if not os.path.exists(RatingsDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + RatingsDir())
        os.makedirs(RatingsDir())
    if not os.path.exists(KeyStoreDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + KeyStoreDir())
        os.makedirs(KeyStoreDir())
    if not os.path.exists(ChatChannelsDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + ChatChannelsDir())
        os.makedirs(ChatChannelsDir())
    if not os.path.exists(ChatHistoryDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + ChatHistoryDir())
        os.makedirs(ChatHistoryDir())


#     if not os.path.exists(BlockchainDir()):
#         if _Debug:
#             lg.out(_DebugLevel, 'settings.init want to create folder: ' + BlockchainDir())
#         os.makedirs(BlockchainDir())


def _checkCustomDirectories():
    """
    Check existance of user configurable folders.
    """
    if config.conf().getString('paths/customers', '') == '':
        config.conf().setString('paths/customers', DefaultCustomersDir())
    if not os.path.exists(getCustomersFilesDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + getCustomersFilesDir())
        os.makedirs(getCustomersFilesDir())
    if config.conf().getString('paths/backups', '') == '':
        config.conf().setString('paths/backups', DefaultBackupsDBDir())
    if not os.path.exists(getLocalBackupsDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + getLocalBackupsDir())
        os.makedirs(getLocalBackupsDir())
    if config.conf().getString('paths/receipts', '') == '':
        config.conf().setString('paths/receipts', DefaultReceiptsDir())
    if not os.path.exists(getReceiptsDir()):
        if _Debug:
            lg.out(_DebugLevel, 'settings.init want to create folder: ' + getReceiptsDir())
        os.makedirs(getReceiptsDir())
    if config.conf().getString('paths/restore', '') == '':
        config.conf().setString('paths/restore', DefaultRestoreDir())


#-------------------------------------------------------------------------------


def main():
    lg.set_debug_level(24)
    init()
    try:
        inp = sys.argv[1].rstrip('/')
    except:
        print('wrong input')
        return
    if not config.conf().exist(inp):
        print('not exist')
        return
    if not config.conf().hasChilds(inp):
        print(inp, config.conf().getData(inp))
        return
    for child in config.conf().listEntries(inp):
        if config.conf().hasChilds(child):
            print(child, config.conf().listEntries(child))
        else:
            print(child, config.conf().getData(child))


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



# Server implementation for a "Sum" command which adds two integers
from __future__ import absolute_import
from twisted.protocols import amp
from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.protocol import Factory


class Sum(amp.Command):
    # normally shared by client and server code
    arguments = [(b'a', amp.Integer()), (b'b', amp.Integer())]
    response = [(b'total', amp.Integer())]


class Protocol(amp.AMP):
    @Sum.responder
    def sum(self, a, b):
        return {'total': a + b}


pf = Factory()
pf.protocol = Protocol
reactor.listenTCP(1234, pf)  # listen on port 1234
reactor.run()



#!/usr/bin/python
# known_nodes.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (known_nodes.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: known_nodes
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

import re

#------------------------------------------------------------------------------


def network_info():
    """
    Returns correct and full network info from `default_network.json` file in the repository root.
    If file `~/.bitdust/[network name]/metadata/networkconfig` exists - use it instead.
    """
    from bitdust.main import network_config
    network_config = network_config.read_network_config_file()
    return network_config['service_entangled_dht']


def default_nodes():
    """
    List of DHT nodes currently maintained : (host, UDP port number)
    """
    from bitdust.lib import strng
    dht_seeds = []
    for dht_seed in network_info()['known_nodes']:
        dht_seeds.append((
            strng.to_bin(dht_seed['host']),
            dht_seed['udp_port'],
        ))
    return dht_seeds


def nodes():
    """
    Here is a well known DHT nodes, this is "genesis" network.
    Every new node in the network will first connect to those nodes,
    and then will be routed to some other nodes already registered.

    Right now we have started several BitDust nodes on vps hosting across the world.
    If you willing to support the project and already started your own BitDust node on reliable machine,
    contact us and we will include your address here.
    So other nodes will be able to use your machine to connect to DHT network.

    The load is not big, but as network will grow we will have more machines listed here,
    so all traffic and ownership will be distributed across the world.

    You can override those "genesis" nodes (before you join network first time)
    by configuring list of your preferred DHT nodes (host or IP address) in the program settings:

        api.config_set(
            "services/entangled-dht/known-nodes",
            "firstnode.net:14441,secondmachine.com:1234,123.45.67.89:9999",
        )

    This way you can create your own DHT network, inside BitDust, under your full control.
    """

    from bitdust.main import config
    from bitdust.lib import strng

    try:
        overridden_dht_nodes_str = str(config.conf().getString('services/entangled-dht/known-nodes'))
    except:
        overridden_dht_nodes_str = ''

    if overridden_dht_nodes_str in [
        'genesis',
        'root',
        b'genesis',
        b'root',
    ]:
        # "genesis" node must not connect anywhere
        return []

    if not overridden_dht_nodes_str:
        return default_nodes()

    overridden_dht_nodes = []
    for dht_node_str in re.split('\n|;|,| ', overridden_dht_nodes_str):
        if dht_node_str.strip():
            try:
                dht_node = dht_node_str.strip().split(':')
                dht_node_host = strng.to_bin(dht_node[0].strip())
                dht_node_port = int(dht_node[1].strip())
            except:
                continue
            overridden_dht_nodes.append((
                dht_node_host,
                dht_node_port,
            ))

    if overridden_dht_nodes:
        return overridden_dht_nodes

    return default_nodes()



#!/usr/bin/env python
# accountant_node.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (accountant_node.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: accountant_node.

.. role:: red

BitDust accountant_node() Automat

EVENTS:
    * :red:`accountant-connected`
    * :red:`coin-broadcasted`
    * :red:`coin-not-valid`
    * :red:`coin-verified`
    * :red:`connection-lost`
    * :red:`init`
    * :red:`lookup-failed`
    * :red:`new-coin-mined`
    * :red:`pending-coin`
    * :red:`shutdown`
    * :red:`start`
    * :red:`stop`
    * :red:`timer-1min`
    * :red:`timer-2min`
    * :red:`valid-coins-received`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

import datetime

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime

from bitdust.automats import automat

from bitdust.transport import callback

from bitdust.p2p import p2p_service
from bitdust.p2p import commands

from bitdust.coins import coins_db
from bitdust.coins import coins_io

from bitdust.broadcast import broadcast_service

#------------------------------------------------------------------------------

_AccountantNode = None

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _AccountantNode
    if event is None and not args:
        return _AccountantNode
    if _AccountantNode is None:
        # set automat name and starting state here
        _AccountantNode = AccountantNode('accountant_node', 'AT_STARTUP', _DebugLevel, _Debug)
    if event is not None:
        _AccountantNode.automat(event, *args, **kwargs)
    return _AccountantNode


#------------------------------------------------------------------------------


class AccountantNode(automat.Automat):
    """
    This class implements all the functionality of the ``accountant_node()``
    state machine.
    """
    timers = {
        'timer-2min': (120, ['ACCOUNTANTS?']),
        'timer-1min': (60, ['READ_COINS']),
    }

    def init(self):
        """
        Method to initialize additional variables and flags at creation phase
        of accountant_node() machine.
        """
        self.connected_accountants = []
        self.min_accountants_connected = 1  # TODO: read from settings
        self.max_accountants_connected = 1  # TODO: read from settings
        self.max_coins_per_packet = 100  # TODO: read from settings
        self.download_offset = datetime.datetime(2016, 1, 1)
        self.download_limit = 100
        self.pending_coins = []
        self.current_coin = None

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python
        <https://bitdust.io/visio2python/>`_ tool.
        """
        #---READY---
        if self.state == 'READY':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'connection-lost' or event == 'stop':
                self.state = 'OFFLINE'
            elif event == 'new-coin-mined':
                self.doPushCoin(*args, **kwargs)
            elif event == 'valid-coins-received':
                self.doWriteCoins(*args, **kwargs)
            elif event == 'pending-coin':
                self.state = 'VALID_COIN?'
                self.doPullCoin(*args, **kwargs)
                self.doVerifyCoin(*args, **kwargs)
        #---READ_COINS---
        elif self.state == 'READ_COINS':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'accountant-connected':
                self.doAddAccountant(*args, **kwargs)
                self.doRetrieveCoins(*args, **kwargs)
            elif event == 'stop' or (event == 'timer-1min' and not self.isAnyCoinsReceived(*args, **kwargs)):
                self.state = 'OFFLINE'
            elif event == 'new-coin-mined':
                self.doPushCoin(*args, **kwargs)
            elif event == 'valid-coins-received' and self.isMoreCoins(*args, **kwargs):
                self.doWriteCoins(*args, **kwargs)
                self.doRetrieveCoins(*args, **kwargs)
            elif event == 'valid-coins-received' and not self.isMoreCoins(*args, **kwargs):
                self.state = 'READY'
                self.doWriteCoins(*args, **kwargs)
                self.doCheckPendingCoins(*args, **kwargs)
        #---AT_STARTUP---
        elif self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'OFFLINE'
                self.doInit(*args, **kwargs)
        #---VALID_COIN?---
        elif self.state == 'VALID_COIN?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'OFFLINE'
            elif event == 'coin-not-valid':
                self.state = 'READY'
                self.doCheckPendingCoins(*args, **kwargs)
            elif event == 'accountant-connected':
                self.doAddAccountant(*args, **kwargs)
            elif event == 'coin-verified':
                self.state = 'WRITE_COIN!'
                self.doWriteCoin(*args, **kwargs)
                self.doBroadcastCoin(*args, **kwargs)
            elif event == 'valid-coins-received':
                self.doWriteCoins(*args, **kwargs)
            elif event == 'new-coin-mined':
                self.doPushCoin(*args, **kwargs)
        #---WRITE_COIN!---
        elif self.state == 'WRITE_COIN!':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'stop':
                self.state = 'OFFLINE'
            elif event == 'coin-broadcasted':
                self.state = 'READY'
                self.doCheckPendingCoins(*args, **kwargs)
            elif event == 'accountant-connected':
                self.doAddAccountant(*args, **kwargs)
            elif event == 'valid-coins-received':
                self.doWriteCoins(*args, **kwargs)
            elif event == 'new-coin-mined':
                self.doPushCoin(*args, **kwargs)
        #---OFFLINE---
        elif self.state == 'OFFLINE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'start':
                self.state = 'ACCOUNTANTS?'
                self.Attempts = 0
                self.doLookupAccountants(*args, **kwargs)
            elif event == 'accountant-connected':
                self.state = 'ACCOUNTANTS?'
                self.doAddAccountant(*args, **kwargs)
                self.Attempts = 2
                self.doLookupAccountants(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---ACCOUNTANTS?---
        elif self.state == 'ACCOUNTANTS?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'accountant-connected' and self.isMoreNeeded(*args, **kwargs):
                self.doAddAccountant(*args, **kwargs)
                self.doLookupAccountants(*args, **kwargs)
            elif event == 'accountant-connected' and not self.isMoreNeeded(*args, **kwargs):
                self.state = 'READ_COINS'
                self.doRetrieveCoins(*args, **kwargs)
            elif event == 'lookup-failed' and self.Attempts < 5 and self.isAnyAccountants(*args, **kwargs):
                self.Attempts += 1
                self.doLookupAccountants(*args, **kwargs)
            elif (event == 'lookup-failed' and (self.Attempts >= 5 or not self.isAnyAccountants(*args, **kwargs))) or event == 'timer-2min':
                self.state = 'OFFLINE'
        return None

    def isAnyAccountants(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.connected_accountants) > 0

    def isMoreNeeded(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.connected_accountants) < self.max_accountants_connected

    def isMoreCoins(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(*args, **kwargs) > 0

    def isAnyCoinsReceived(self, *args, **kwargs):
        """
        Condition method.
        """

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        callback.append_inbox_callback(self._on_inbox_packet)

    def doLookupAccountants(self, *args, **kwargs):
        """
        Action method.
        """
        from bitdust.coins import accountants_finder
        accountants_finder.A('start', (self.automat, {
            'action': 'join',
        }))

    def doAddAccountant(self, *args, **kwargs):
        """
        Action method.
        """
        if args[0] in self.connected_accountants:
            if _Debug:
                lg.out(_DebugLevel, 'accountant_node.doAddAccountant SKIP, %s already connected, skip' % args[0])
            return
        self.connected_accountants.append(args[0])
        if _Debug:
            lg.out(_DebugLevel, 'accountant_node.doAddAccountant NEW %s connected, %d total accountants' % (args[0], len(self.connected_accountants)))

    def doRetrieveCoins(self, *args, **kwargs):
        """
        Action method.
        """
        query = {
            'method': 'get_many',
            'index': 'time_mined',
            'limit': self.download_limit,
            'start': utime.datetime_to_sec1970(self.download_offset),
            'end': utime.utcnow_to_sec1970(),
        }
        for idurl in self.connected_accountants:
            p2p_service.SendRetrieveCoin(idurl, query)

    def doWriteCoins(self, *args, **kwargs):
        """
        Action method.
        """
        for coin in args[0]:
            if not coins_db.exist(coin):
                coins_db.insert(coin)
            if coin['tm'] > utime.datetime_to_sec1970(self.download_offset):
                self.download_offset = utime.sec1970_to_datetime_utc(coin['tm'])

    def doPushCoin(self, *args, **kwargs):
        """
        Action method.
        """
        self.pending_coins.append(args[0])

    def doPullCoin(self, *args, **kwargs):
        """
        Action method.
        """
        self.current_coin = self.pending_coins.pop(0)

    def doCheckPendingCoins(self, *args, **kwargs):
        """
        Action method.
        """
        if len(self.pending_coins) > 0:
            self.automat('pending-coin')

    def doVerifyCoin(self, *args, **kwargs):
        """
        Action method.
        """
        if not coins_io.validate_coin(self.current_coin):
            self.current_coin = None
            self.automat('coin-not-valid')
            return
        d = self._verify_coin(self.current_coin)
        d.addCallback(lambda coin: self.automat('coin-verified'))
        d.addErrback(lambda err: self.automat('coin-not-valid'))

    def doBroadcastCoin(self, *args, **kwargs):
        """
        Action method.
        """
        broadcast_service.send_broadcast_message({
            'type': 'coin',
            'data': args[0],
        })

    def doWriteCoin(self, *args, **kwargs):
        """
        Action method.
        """
        coins_db.insert(args[0])

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.pending_coins = None
        self.connected_accountants = None
        callback.remove_inbox_callback(self._on_inbox_packet)
        self.destroy()
        global _AccountantNode
        del _AccountantNode
        _AccountantNode = None

    #------------------------------------------------------------------------------

    def _verify_coin(self, acoin):
        # TODO:
        # run coin verification process
        # may require to communicate with other accountants
        d = Deferred()
        d.callback(acoin)
        return d

    def _on_command_retrieve_coin(self, newpacket, info):
        query_j = coins_io.read_query_from_packet(newpacket)
        if not query_j:
            p2p_service.SendFail(newpacket, 'incorrect query received')
            return False
        coins, error = coins_db.query_json(query_j)
        if error:
            p2p_service.SendFail(newpacket, error)
            return False
        result_coins = []
        for coin in coins:
            result_coins.append(coin)
            if len(result_coins) > self.max_coins_per_packet:
                result_coins.append(None)
                break
        if not result_coins:
            p2p_service.SendFail(newpacket, 'no coins found')
            return False
        p2p_service.SendCoin(newpacket.CreatorID, result_coins, packet_id=newpacket.PacketID)
        return True

    def _on_command_coin(self, newpacket, info):
        coins_list = coins_io.read_coins_from_packet(newpacket)
        if not coins_list:
            p2p_service.SendFail(newpacket, 'failed to read coins from packet')
            return True
        if coins_list[-1] is None:
            # TODO: partial result, request more coins
            lg.warn('partial response, more coins were found but not transferred')
            del coins_list[-1]
        if len(coins_list) == 1:
            acoin = coins_list[0]
            if not coins_io.validate_coin(acoin):
                p2p_service.SendFail(newpacket, 'coin validation failed')
                return True
            if not coins_io.verify_coin(acoin):
                p2p_service.SendFail(newpacket, 'coin verification failed')
                return True
            if coins_db.exist(acoin):
                self.automat(
                    'valid-coins-received',
                    [
                        acoin,
                    ],
                )
            else:
                self.automat('new-coin-mined', acoin)
            return True
        valid_coins = []
        for acoin in coins_list:
            if not coins_io.validate_coin(acoin):
                continue
            if not coins_io.verify_coin(acoin):
                continue
            valid_coins.append(acoin)
        if len(valid_coins) == len(coins_list):
            self.automat('valid-coins-received', valid_coins)
        else:
            p2p_service.SendFail(newpacket, 'some non-valid coins received')
        return True

    def _on_inbox_packet(self, newpacket, info, status, error_message):
        if _Debug:
            lg.out(_DebugLevel, 'accountant_node._on_inbox_packet %r' % newpacket)
        if newpacket.Command == commands.RetrieveCoin():
            return self._on_command_retrieve_coin(newpacket, info)
        if newpacket.Command == commands.Coin():
            return self._on_command_coin(newpacket, info)
        return False



import time
import sys
"""
Call send_nogui.py (edit not to require manual confirmation)
for each address to pay.

rewards.csv is to be in the same dir.
format is one payout per line, comma separated, address,amount,extra

```25125e9bb305fafd51ceb2858d355f77da99550b933ec0923cd156ff,1310.4750655411829,5111
8f2d03c817c3d36a864c99a27f6b6179eb1898a631bc007a7e0ffa39,603.0595488461871,2352
0fc9b60126b8b5be3ab990eea6f184b02c1c0c5352709d023256ca58,459.7303448474547,1793```

Amount really sent will be reduced by the tx fee, 0.01

NO SAFETY there, be sure what you do.

The node has to be running with mempool on disk, not on ram or send_nogui does not work!!!
"""

import argparse
import os

__version__ = '0.0.1'

SEND_PATH = 'send_nogui_noconf.py'  # path to modified send_no_gui.py in the Bismuth Dir.
# That node has to be running with mempool on disk, not on ram!!!

PYTHON_EXECUTABLE = 'python3'

parser = argparse.ArgumentParser(description='Bismuth Batch reward sender')
# parser.add_argument("-v", "--verbose", action="count", default=False, help='Be verbose.')
parser.add_argument('-y', '--yes', action='count', default=False, help='Do send')
parser.add_argument('-w', '--wallet', help='Path to wallet, use quotation marks')
args = parser.parse_args()

print(sys.argv[3])

total = 0
nb = 0
for line in open('rewards.csv', 'r'):
    data = line.strip().split(',')
    print(data)
    if len(data) > 1:
        try:
            total += float(data[1])
            data[1] = float(data[1]) - 0.01
            command = f'{PYTHON_EXECUTABLE} {SEND_PATH} {data[1]} {data[0]} {None} {None} {sys.argv[3]} '  #arguments are passed here
            if args.yes:
                print(f'Running: {command} tx')
                os.system(command)
            else:
                print(f"Check: {command}, didn't you forget the magic word?")
                sys.exit(0)
            nb += 1
            time.sleep(1)
        except Exception as e:
            print(e)

print(f'{nb} Transactions, {total} $BIS total.')



#!/usr/bin/env python
# contact.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

import six

_Debug = False


class Contact(object):
    """
    Encapsulation for remote contact.

    This class contains information on a single remote contact, and also
    provides a direct RPC API to the remote node which it represents
    """
    def __init__(self, id, ipAddress, udpPort, networkProtocol, firstComm=0):
        self.id = id
        self.address = ipAddress
        self.port = udpPort
        self._networkProtocol = networkProtocol
        self.commTime = firstComm

    def __eq__(self, other):
        if isinstance(other, Contact):
            return self.id == other.id
        elif isinstance(other, six.string_types):
            return self.id == other
        else:
            return False

    def __ne__(self, other):
        if isinstance(other, Contact):
            return self.id != other.id
        elif isinstance(other, six.string_types):
            return self.id != other
        else:
            return True

    def __repr__(self, *args, **kwargs):
        return str(self)

    def __str__(self):
        return '<%s at %r:%d>' % (self.id[:6], self.address, self.port)

    def __getattr__(self, name):
        """
        This override allows the host node to call a method of the remote node
        (i.e. this contact) as if it was a local function.

        For instance, if C{remoteNode} is a instance of C{Contact}, the
        following will result in C{remoteNode}'s C{test()} method to be
        called with argument C{123}::
         remoteNode.test(123)

        Such a RPC method call will return a Deferred, which will callback
        when the contact responds with the result (or an error occurs).
        This happens via this contact's C{_networkProtocol} object (i.e. the
        host Node's C{_protocol} object).
        """
        def _sendRPC(*args, **kwargs):
            return self._networkProtocol.sendRPC(self, name, args, **kwargs)

        return _sendRPC


class LayeredContact(Contact):
    def __init__(self, id, ipAddress, udpPort, networkProtocol, firstComm=0, layerID=0):
        self.layerID = layerID
        super(LayeredContact, self).__init__(id, ipAddress, udpPort, networkProtocol, firstComm)

    def __str__(self):
        return '<Contact(%d) %s at %r:%d>' % (self.layerID, self.id[:6], self.address, self.port)

    def __getattr__(self, name):
        def _sendRPC(*args, **kwargs):
            kwargs['layerID'] = self.layerID
            if _Debug:
                print('[DHT CONTACT]: %r sending RPC %r args=%r kwargs=%r' % (self, name, args, kwargs))
            ret = self._networkProtocol.sendRPC(self, name, args, **kwargs)
            if _Debug:
                print('[DHT CONTACT]: %r RPC result is %r' % (self, ret))
            return ret

        return _sendRPC



#!/usr/bin/python
# restore_monitor.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (restore_monitor.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
"""
.. module:: restore_monitor.

Manages currently restoring backups.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import events

from bitdust.lib import misc

from bitdust.system import tmpfile

from bitdust.storage import backup_tar
from bitdust.storage import backup_matrix
from bitdust.storage import backup_control

from bitdust.userid import global_id

#------------------------------------------------------------------------------

_WorkingBackupIDs = {}
_WorkingRestoreProgress = {}

#------------------------------------------------------------------------------

OnRestorePacketFunc = None
OnRestoreDoneFunc = None
OnRestoreBlockFunc = None

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'restore_monitor.init')


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'restore_monitor.shutdown')


#------------------------------------------------------------------------------


def block_restored_callback(backupID, block):
    global OnRestoreBlockFunc
    if OnRestoreBlockFunc is not None:
        OnRestoreBlockFunc(backupID, block)


def packet_in_callback(backupID, newpacket):
    global _WorkingRestoreProgress
    global OnRestorePacketFunc
    SupplierNumber = newpacket.SupplierNumber()
    if _Debug:
        lg.out(_DebugLevel, 'restore_monitor.packet_in_callback %s from supplier %s' % (backupID, SupplierNumber))
    # want to count the data we restoring
    if SupplierNumber not in list(_WorkingRestoreProgress[backupID].keys()):
        _WorkingRestoreProgress[backupID][SupplierNumber] = 0
    _WorkingRestoreProgress[backupID][SupplierNumber] += len(newpacket.Payload)
    packetID = global_id.CanonicalID(newpacket.PacketID)
    backup_matrix.LocalFileReport(packetID)
    if OnRestorePacketFunc is not None:
        OnRestorePacketFunc(backupID, SupplierNumber, newpacket)


def extract_done(retcode, backupID, source_filename, output_location, callback_method):
    lg.info('extract success of %s with result : %s' % (backupID, str(retcode)))
    global OnRestoreDoneFunc
    _WorkingBackupIDs.pop(backupID, None)
    _WorkingRestoreProgress.pop(backupID, None)
    tmpfile.throw_out(source_filename, 'file extracted')
    if OnRestoreDoneFunc is not None:
        OnRestoreDoneFunc(backupID, 'restore done')
    if callback_method:
        try:
            callback_method(backupID, 'restore done')
        except:
            lg.exc()
    events.send('restore-done', data=dict(
        backup_id=backupID,
        output_location=output_location,
    ))
    return retcode


def extract_failed(err, backupID, source_filename, output_location, callback_method):
    lg.err('extract failed of %s with: %s' % (backupID, str(err)))
    global OnRestoreDoneFunc
    _WorkingBackupIDs.pop(backupID, None)
    _WorkingRestoreProgress.pop(backupID, None)
    tmpfile.throw_out(source_filename, 'file extract failed')
    if OnRestoreDoneFunc is not None:
        OnRestoreDoneFunc(backupID, 'extract failed')
    if callback_method:
        try:
            callback_method(backupID, 'extract failed')
        except:
            lg.exc()
    events.send('restore-failed', data=dict(
        backup_id=backupID,
        output_location=output_location,
        reason='extracting file failed',
        error=str(err),
    ))
    return err


def restore_done(result, backupID, outfd, tarfilename, outputlocation, callback_method):
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    global OnRestoreDoneFunc
    if _Debug:
        lg.args(_DebugLevel, result=result, bid=backupID, tar=tarfilename, out=outputlocation)
    if result == 'done':
        lg.info('restore success of %s with result=%s' % (backupID, result))
    else:
        lg.err('restore failed of %s with result=%s' % (backupID, result))
    try:
        os.close(outfd)
    except:
        lg.exc()
    if result == 'done':
        d = backup_tar.extracttar_thread(tarfilename, outputlocation)
        d.addCallback(extract_done, backupID, tarfilename, outputlocation, callback_method)
        d.addErrback(extract_failed, backupID, tarfilename, outputlocation, callback_method)
        return d
    _WorkingBackupIDs.pop(backupID, None)
    _WorkingRestoreProgress.pop(backupID, None)
    tmpfile.throw_out(tarfilename, 'restore ' + result)
    if OnRestoreDoneFunc is not None:
        OnRestoreDoneFunc(backupID, result)
    if callback_method:
        try:
            callback_method(backupID, result)
        except:
            lg.exc()
    return result


#------------------------------------------------------------------------------


def Start(backupID, outputLocation, callback=None, keyID=None):
    if _Debug:
        lg.out(_DebugLevel, 'restore_monitor.Start %s to %s' % (backupID, outputLocation))
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    if backupID in list(_WorkingBackupIDs.keys()):
        return _WorkingBackupIDs[backupID]
    alias = backupID.split('$')[0]
    outfd, outfilename = tmpfile.make(
        'restore',
        extension='.tar.gz',
        prefix=alias + '_',
    )
    from bitdust.storage import restore_worker
    r = restore_worker.RestoreWorker(backupID, outfd, KeyID=keyID)
    r.MyDeferred.addCallback(restore_done, backupID, outfd, outfilename, outputLocation, callback)
    r.set_block_restored_callback(block_restored_callback)
    r.set_packet_in_callback(packet_in_callback)
    _WorkingBackupIDs[backupID] = r
    _WorkingRestoreProgress[backupID] = {}
    r.automat('init')
    return r


def Abort(backupID):
    global _WorkingBackupIDs
    global _WorkingRestoreProgress
    if backupID not in list(_WorkingBackupIDs.keys()):
        lg.warn('%s not found in working list' % backupID)
        return False
    r = _WorkingBackupIDs[backupID]
    r.automat('abort', 'abort')
    if _Debug:
        lg.out(_DebugLevel, 'restore_monitor.Abort %s' % backupID)
    return True


#------------------------------------------------------------------------------


def GetWorkingIDs():
    global _WorkingBackupIDs
    return list(_WorkingBackupIDs.keys())


def GetWorkingObjects():
    global _WorkingBackupIDs
    return list(_WorkingBackupIDs.values())


def IsWorking(backupID):
    global _WorkingBackupIDs
    return backupID in list(_WorkingBackupIDs.keys())


def GetProgress(backupID):
    global _WorkingRestoreProgress
    return _WorkingRestoreProgress.get(backupID, {})


def GetWorkingRestoreObject(backupID):
    global _WorkingBackupIDs
    return _WorkingBackupIDs.get(backupID, None)


def FindWorking(pathID=None, customer=None):
    global _WorkingBackupIDs
    if pathID:
        pathID = global_id.CanonicalID(pathID)
    result = set()
    for backupID in _WorkingBackupIDs:
        if pathID:
            if backupID.count(pathID):
                result.add(backupID)
        if customer:
            if backupID.count(customer + ':'):
                result.add(backupID)
    return list(result)


def GetBackupStatusInfo(backupID, item_info, item_name, parent_path_existed=None):
    _, percent, weakBlock, weakPercent = backup_matrix.GetBackupRemoteStats(backupID)
    totalNumberOfFiles, maxBlockNum, statsArray = backup_matrix.GetBackupStats(backupID)
    ret = {
        'state': 'ready',
        'delivered': misc.percent2string(percent),
        'reliable': misc.percent2string(weakPercent),
        'fragments': totalNumberOfFiles,
        'weak_block': weakBlock,
        'max_block': maxBlockNum,
        'suppliers': [{
            'stored': misc.percent2string(i[0]),
            'fragments': i[1],
        } for i in statsArray],
    }
    backupObj = backup_control.GetRunningBackupObject(backupID)
    if backupObj:
        ret['state'] = 'uploading'
        ret['progress'] = misc.percent2string(backupObj.progress())
        return ret
    elif IsWorking(backupID):
        restoreObj = GetWorkingRestoreObject(backupID)
        if restoreObj:
            maxBlockNum = backup_matrix.GetKnownMaxBlockNum(backupID)
            currentBlock = max(0, restoreObj.block_number)
            percent = 0.0
            if maxBlockNum > 0:
                percent = 100.0*currentBlock/maxBlockNum
            ret['state'] = 'downloading'
            ret['progress'] = misc.percent2string(percent)
            return ret
    return ret



import sqlite3, time

reward_10 = 14.99999
row_init = 10
row_end = 660488
row = row_init
reward = reward_10

conn = sqlite3.connect('static/ledger.db')
conn.text_factory = str
c = conn.cursor()

c.execute("DELETE FROM transactions WHERE address = 'Development Reward'")
conn.commit()

while row <= row_end:
    print(row)
    timestamp = int(time.time())

    c.execute('INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)', (0, timestamp, 'Development Reward', '4edadac9093d9326ee4b17f869b14f1a2534f96f9c5d7b48dc9acaed', reward, '0', '0', '0', '0', '0', '0', row))

    row = row + 10
    reward = reward - 0.00001
conn.commit()
